"""Testing DTI."""

import numpy as np
import numpy.testing as npt

import dipy.core.gradients as grad
import dipy.core.sphere as dps
from dipy.core.subdivide_octahedron import create_unit_sphere
from dipy.data import dsi_voxels, get_fnames, get_sphere
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti_data
import dipy.reconst.dti as dti
from dipy.reconst.dti import (
    MIN_POSITIVE_SIGNAL,
    TensorModel,
    _decompose_tensor_nan,
    axial_diffusivity,
    color_fa,
    decompose_tensor,
    fractional_anisotropy,
    from_lower_triangular,
    geodesic_anisotropy,
    linearity,
    lower_triangular,
    mean_diffusivity,
    mode,
    ols_fit_tensor,
    planarity,
    radial_diffusivity,
    sphericity,
    trace,
    wls_fit_tensor,
)
from dipy.reconst.weights_method import (
    weights_method_nlls_m_est,
    weights_method_wls_m_est,
)
from dipy.sims.voxel import single_tensor
from dipy.testing.decorators import set_random_number_generator


def test_roll_evals():
    # Just making sure this never passes through
    weird_evals = np.array([1, 0.5])
    npt.assert_raises(ValueError, dti._roll_evals, weird_evals)


@set_random_number_generator()
def test_tensor_algebra(rng):
    # Test that the computation of tensor determinant and norm is correct
    test_arr = rng.random((10, 3, 3))
    t_det = dti.determinant(test_arr)
    t_norm = dti.norm(test_arr)
    for i, x in enumerate(test_arr):
        npt.assert_almost_equal(np.linalg.det(x), t_det[i])
        npt.assert_almost_equal(np.linalg.norm(x), t_norm[i])


def test_odf_with_zeros():
    fdata, fbval, fbvec = get_fnames(name="small_25")
    gtab = grad.gradient_table(fbval, bvecs=fbvec)
    data = load_nifti_data(fdata)
    dm = dti.TensorModel(gtab)
    df = dm.fit(data)
    df.evals[0, 0, 0] = np.array([0, 0, 0])
    sphere = create_unit_sphere(recursion_level=4)
    odf = df.odf(sphere)
    npt.assert_equal(odf[0, 0, 0], np.zeros(sphere.vertices.shape[0]))


def test_mode_with_isotropic():
    # mode involves a division by norm, so may be problematic for isotropic
    # voxels. In the above test, 4 voxels are produced with isotropic tensors
    # in indexes [0, 0] and [0, 1] in which norm should give mode 0. Voxels
    # with indexes [1, 0] and [1, 1] should give mode of 1 and -1 respectively.
    q_form = np.zeros((2, 2, 3, 3))
    q_form[0, 1, 0, 0] = 1
    q_form[0, 1, 1, 1] = 1
    q_form[0, 1, 2, 2] = 1
    q_form[1, 0, 0, 0] = 1
    q_form[1, 0, 1, 1] = 1
    q_form[1, 0, 2, 2] = 2
    q_form[1, 1, 0, 0] = 1
    q_form[1, 1, 1, 1] = 2
    q_form[1, 1, 2, 2] = 2
    npt.assert_array_almost_equal(mode(q_form), np.array([[0, 0], [1, -1]]))


def test_tensor_model():
    fdata, fbval, fbvec = get_fnames(name="small_25")
    data1 = load_nifti_data(fdata)
    gtab1 = grad.gradient_table(fbval, bvecs=fbvec)
    data2, gtab2 = dsi_voxels()
    for data, gtab in zip([data1, data2], [gtab1, gtab2]):
        dm = dti.TensorModel(gtab, fit_method="LS")
        dtifit = dm.fit(data[0, 0, 0])
        npt.assert_equal(dtifit.fa < 0.9, True)
        dm = dti.TensorModel(gtab, fit_method="WLS")
        dtifit = dm.fit(data[0, 0, 0])
        npt.assert_equal(dtifit.fa < 0.9, True)
        npt.assert_equal(dtifit.fa > 0, True)
        sphere = create_unit_sphere(recursion_level=4)
        npt.assert_equal(len(dtifit.odf(sphere)), len(sphere.vertices))
        # Check that the multivoxel case works:
        dtifit = dm.fit(data)

        # Check that it works on signal that has already been normalized to S0:
        dm_to_relative = dti.TensorModel(gtab)
        if np.any(gtab.b0s_mask):
            relative_data = data[0, 0, 0] / np.mean(data[0, 0, 0, gtab.b0s_mask])

            dtifit_to_relative = dm_to_relative.fit(relative_data)
            npt.assert_almost_equal(
                dtifit.fa[0, 0, 0], dtifit_to_relative.fa, decimal=3
            )

    # And smoke-test that all these operations return sensibly-shaped arrays:
    npt.assert_equal(dtifit.fa.shape, data.shape[:3])
    npt.assert_equal(dtifit.ad.shape, data.shape[:3])
    npt.assert_equal(dtifit.md.shape, data.shape[:3])
    npt.assert_equal(dtifit.rd.shape, data.shape[:3])
    npt.assert_equal(dtifit.trace.shape, data.shape[:3])
    npt.assert_equal(dtifit.mode.shape, data.shape[:3])
    npt.assert_equal(dtifit.linearity.shape, data.shape[:3])
    npt.assert_equal(dtifit.planarity.shape, data.shape[:3])
    npt.assert_equal(dtifit.sphericity.shape, data.shape[:3])

    # Test for the shape of the mask
    npt.assert_raises(ValueError, dm.fit, np.ones((10, 10, 3)), mask=np.ones((3, 3)))

    # Make some synthetic data
    b0 = 1000.0
    bvals, bvecs = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table_from_bvals_bvecs(bvals, bvecs)
    # The first b value is 0., so we take the second one:
    B = bvals[1]
    # Scale the eigenvalues and tensor by the B value so the units match
    D = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 1.0, -np.log(b0) * B]) / B
    evals = np.array([2.0, 1.0, 0.0]) / B
    md = evals.mean()
    tensor = from_lower_triangular(D)
    A_squiggle = tensor - (1 / 3.0) * np.trace(tensor) * np.eye(3)
    mode = 3 * np.sqrt(6) * np.linalg.det(A_squiggle / np.linalg.norm(A_squiggle))
    evals_eigh, evecs_eigh = np.linalg.eigh(tensor)
    # Sort according to eigen-value from large to small:
    evecs = evecs_eigh[:, np.argsort(evals_eigh)[::-1]]
    # Check that eigenvalues and eigenvectors are properly sorted through
    # that previous operation:
    for i in range(3):
        npt.assert_array_almost_equal(
            np.dot(tensor, evecs[:, i]), evals[i] * evecs[:, i]
        )
    # Design Matrix
    X = dti.design_matrix(gtab)
    # Signals
    Y = np.exp(np.dot(X, D))
    npt.assert_almost_equal(Y[0], b0)
    Y.shape = (-1,) + Y.shape

    # Test fitting with different methods:
    for fit_method in ["OLS", "WLS", "NLLS"]:
        tensor_model = dti.TensorModel(gtab, fit_method=fit_method, return_S0_hat=True)

        tensor_fit = tensor_model.fit(Y)
        assert tensor_fit.model is tensor_model
        npt.assert_equal(tensor_fit.shape, Y.shape[:-1])
        npt.assert_array_almost_equal(tensor_fit.evals[0], evals)
        npt.assert_array_almost_equal(tensor_fit.S0_hat, b0, decimal=3)
        # Test that the eigenvectors are correct, one-by-one:
        for i in range(3):
            # Eigenvectors have intrinsic sign ambiguity
            # (see
            # http://prod.sandia.gov/techlib/access-control.cgi/2007/076422.pdf)
            # so we need to allow for sign flips. One of the following should
            # always be true:
            npt.assert_(
                np.all(np.abs(tensor_fit.evecs[0][:, i] - evecs[:, i]) < 10e-6)
                or np.all(np.abs(-tensor_fit.evecs[0][:, i] - evecs[:, i]) < 10e-6)
            )
            # We set a fixed tolerance of 10e-6, similar to array_almost_equal

        err_msg = "Calculation of tensor from Y does not compare to "
        err_msg += "analytical solution"
        npt.assert_array_almost_equal(
            tensor_fit.quadratic_form[0], tensor, err_msg=err_msg
        )

        npt.assert_almost_equal(tensor_fit.md[0], md)
        npt.assert_array_almost_equal(tensor_fit.mode, mode, decimal=5)
        npt.assert_equal(tensor_fit.directions.shape[-2], 1)
        npt.assert_equal(tensor_fit.directions.shape[-1], 3)

    # Test error-handling:
    npt.assert_raises(ValueError, dti.TensorModel, gtab, fit_method="crazy_method")
    npt.assert_warns(UserWarning, dti.TensorModel, gtab, fit_method="NLLS", step=1e4)
    with npt.assert_warns(UserWarning):
        model = dti.TensorModel(gtab, fit_method="NLS", step=1e4)
        npt.assert_equal(model.kwargs.get("step", None), None)

    # Test custom fit tensor method
    try:
        model = dti.TensorModel(gtab, fit_method=lambda *args, **kwargs: 42)
        fit = model.fit_method()
    except Exception as exc:
        raise AssertionError(
            f"TensorModel should accept custom fit methods: {exc}"
        ) from exc
    assert fit == 42, f"Custom fit method for TensorModel returned {fit}."

    # Test multi-voxel data
    data = np.zeros((3, Y.shape[1]))
    # Normal voxel
    data[0] = Y
    # High diffusion voxel, all diffusing weighted signal equal to zero
    data[1, gtab.b0s_mask] = b0
    data[1, ~gtab.b0s_mask] = 0
    # Masked voxel, all data set to zero
    data[2] = 0.0

    tensor_model = dti.TensorModel(gtab)
    fit = tensor_model.fit(data)
    npt.assert_array_almost_equal(fit[0].evals, evals)

    # Return S0_test
    tensor_model = dti.TensorModel(gtab, return_S0_hat=True)
    fit = tensor_model.fit(data)
    npt.assert_array_almost_equal(fit[0].evals, evals)
    npt.assert_array_almost_equal(fit[0].S0_hat, b0)

    # Evals should be high for high diffusion voxel
    assert all(fit[1].evals > evals[0] * 0.9)

    # Evals should be zero where data is masked
    npt.assert_array_almost_equal(fit[2].evals, 0.0)


def test_indexing_on_tensor_fit():
    params = np.zeros([2, 3, 4, 12])
    fit = dti.TensorFit(None, params)

    # Should return a TensorFit of appropriate shape
    npt.assert_equal(fit.shape, (2, 3, 4))
    fit1 = fit[0]
    npt.assert_equal(fit1.shape, (3, 4))
    npt.assert_equal(type(fit1), dti.TensorFit)
    fit1 = fit[0, 0, 0]
    npt.assert_equal(fit1.shape, ())
    npt.assert_equal(type(fit1), dti.TensorFit)
    fit1 = fit[[0], slice(None)]
    npt.assert_equal(fit1.shape, (1, 3, 4))
    npt.assert_equal(type(fit1), dti.TensorFit)

    # Should raise an index error if too many indices are passed
    npt.assert_raises(IndexError, fit.__getitem__, (0, 0, 0, 0))


def test_fa_of_zero():
    evals = np.zeros((4, 3))
    fa = fractional_anisotropy(evals)
    npt.assert_array_equal(fa, 0)


def test_ga_of_zero():
    evals = np.zeros((4, 3))
    ga = geodesic_anisotropy(evals)
    npt.assert_array_equal(ga, 0)


def test_diffusivities():
    psphere = get_sphere(name="symmetric362")
    bvecs = np.concatenate(([[0, 0, 0]], psphere.vertices))
    bvals = np.zeros(len(bvecs)) + 1000
    bvals[0] = 0
    gtab = grad.gradient_table(bvals, bvecs=bvecs)
    mevals = np.array(([0.0015, 0.0003, 0.0001], [0.0015, 0.0003, 0.0003]))
    mevecs = [
        np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        np.array([[0, 0, 1], [0, 1, 0], [1, 0, 0]]),
    ]
    S = single_tensor(gtab, 100, evals=mevals[0], evecs=mevecs[0], snr=None)

    dm = dti.TensorModel(gtab, fit_method="LS")
    dmfit = dm.fit(S)

    md = mean_diffusivity(dmfit.evals)
    Trace = trace(dmfit.evals)
    rd = radial_diffusivity(dmfit.evals)
    ad = axial_diffusivity(dmfit.evals)
    lin = linearity(dmfit.evals)
    plan = planarity(dmfit.evals)
    spher = sphericity(dmfit.evals)

    npt.assert_almost_equal(md, (0.0015 + 0.0003 + 0.0001) / 3)
    npt.assert_almost_equal(Trace, (0.0015 + 0.0003 + 0.0001))
    npt.assert_almost_equal(ad, 0.0015)
    npt.assert_almost_equal(rd, (0.0003 + 0.0001) / 2)
    npt.assert_almost_equal(lin, (0.0015 - 0.0003) / Trace)
    npt.assert_almost_equal(plan, 2 * (0.0003 - 0.0001) / Trace)
    npt.assert_almost_equal(spher, (3 * 0.0001) / Trace)


def test_color_fa():
    data, gtab = dsi_voxels()
    dm = dti.TensorModel(gtab, fit_method="LS")
    dmfit = dm.fit(data)
    fa = fractional_anisotropy(dmfit.evals)

    fa = np.ones((3, 3, 3))
    # evecs should be of shape (fa, 3, 3)
    evecs = np.zeros(fa.shape + (3, 2))
    npt.assert_raises(ValueError, color_fa, fa, evecs)

    evecs = np.zeros(fa.shape + (3, 3))
    evecs[..., :, :] = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    npt.assert_equal(fa.shape, evecs[..., 0, 0].shape)
    npt.assert_equal((3, 3), evecs.shape[-2:])

    # 3D test case
    fa = np.ones((3, 3, 3))
    evecs = np.zeros(fa.shape + (3, 3))
    evecs[..., :, :] = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    cfa = color_fa(fa, evecs)
    cfa_truth = np.array([1, 0, 0])
    true_cfa = np.reshape(np.tile(cfa_truth, 27), [3, 3, 3, 3])

    npt.assert_array_equal(cfa, true_cfa)

    # 2D test case
    fa = np.ones((3, 3))
    evecs = np.zeros(fa.shape + (3, 3))
    evecs[..., :, :] = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    cfa = color_fa(fa, evecs)
    cfa_truth = np.array([1, 0, 0])
    true_cfa = np.reshape(np.tile(cfa_truth, 9), [3, 3, 3])

    npt.assert_array_equal(cfa, true_cfa)

    # 1D test case
    fa = np.ones(3)
    evecs = np.zeros(fa.shape + (3, 3))
    evecs[..., :, :] = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    cfa = color_fa(fa, evecs)
    cfa_truth = np.array([1, 0, 0])
    true_cfa = np.reshape(np.tile(cfa_truth, 3), [3, 3])

    npt.assert_array_equal(cfa, true_cfa)


def test_wls_and_ls_fit():
    """
    Tests the WLS and LS fitting functions to see if they returns the correct
    eigenvalues and eigenvectors.

    Uses data/55dir_grad as the gradient table and 3by3by56.nii
    as the data.

    """

    # Defining Test Voxel (avoid nibabel dependency) ###

    # Recall: D = [Dxx,Dyy,Dzz,Dxy,Dxz,Dyz,log(S_0)] and D ~ 10^-4 mm^2 /s
    b0 = 1000.0
    bval, bvec = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    B = bval[1]
    # Scale the eigenvalues and tensor by the B value so the units match
    D = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 1.0, -np.log(b0) * B]) / B
    evals = np.array([2.0, 1.0, 0.0]) / B
    md = evals.mean()
    tensor = from_lower_triangular(D)
    # Design Matrix
    gtab = grad.gradient_table(bval, bvecs=bvec)
    X = dti.design_matrix(gtab)
    # Signals
    Y = np.exp(np.dot(X, D))
    npt.assert_almost_equal(Y[0], b0)
    Y.shape = (-1,) + Y.shape

    # Testing WLS Fit on single voxel
    # If you do something wonky (passing min_signal<0), you should get an
    # error:
    npt.assert_raises(ValueError, TensorModel, gtab, fit_method="WLS", min_signal=-1)

    # Estimate tensor from test signals
    model = TensorModel(gtab, fit_method="WLS", return_S0_hat=True)
    tensor_est = model.fit(Y)
    npt.assert_equal(tensor_est.shape, Y.shape[:-1])
    npt.assert_array_almost_equal(tensor_est.evals[0], evals)
    npt.assert_array_almost_equal(
        tensor_est.quadratic_form[0],
        tensor,
        err_msg="Calculation of tensor from Y does "
        "not compare to analytical solution",
    )
    npt.assert_almost_equal(tensor_est.md[0], md)
    npt.assert_array_almost_equal(tensor_est.S0_hat[0], b0, decimal=3)

    # Test that we can fit a single voxel's worth of data (a 1d array)
    y = Y[0]
    tensor_est = model.fit(y)
    npt.assert_equal(tensor_est.shape, ())
    npt.assert_array_almost_equal(tensor_est.evals, evals)
    npt.assert_array_almost_equal(tensor_est.quadratic_form, tensor)
    npt.assert_almost_equal(tensor_est.md, md)
    npt.assert_array_almost_equal(tensor_est.lower_triangular(b0=b0), D)

    # Test using fit_method='LS'
    model = TensorModel(gtab, fit_method="LS")
    tensor_est = model.fit(y)
    npt.assert_equal(tensor_est.shape, ())
    npt.assert_array_almost_equal(tensor_est.evals, evals)
    npt.assert_array_almost_equal(tensor_est.quadratic_form, tensor)
    npt.assert_almost_equal(tensor_est.md, md)
    npt.assert_array_almost_equal(tensor_est.lower_triangular(b0=b0), D)
    npt.assert_array_almost_equal(tensor_est.linearity, linearity(evals))
    npt.assert_array_almost_equal(tensor_est.planarity, planarity(evals))
    npt.assert_array_almost_equal(tensor_est.sphericity, sphericity(evals))

    # testing that leverages are returned on request
    for fit_method in ["LS", "WLS"]:
        # Estimate tensor from test signals, not returning leverages
        model = TensorModel(
            gtab, fit_method=fit_method, return_S0_hat=True, return_leverages=False
        )
        tensor_est = model.fit(Y)
        npt.assert_equal(tensor_est.model.extra, {})

        # Estimate tensor from test signals, returning leverages
        model = TensorModel(
            gtab, fit_method=fit_method, return_S0_hat=True, return_leverages=True
        )
        tensor_est = model.fit(Y)
        npt.assert_equal(tensor_est.model.extra["leverages"].shape, Y.shape)

        # test value of leverage is 7 (in this case, for DTI)
        leverages = tensor_est.model.extra["leverages"]
        npt.assert_almost_equal(leverages.sum(axis=1), np.array([7.0]))

    # Test wls given S^2 weights argument, matches default wls
    design_matrix = dti.design_matrix(gtab)
    YN = Y + 10 * np.random.normal(size=Y.shape)  # error or weights irrelevant
    YN[YN < MIN_POSITIVE_SIGNAL] = MIN_POSITIVE_SIGNAL
    # wls calculation
    D_w, _ = wls_fit_tensor(design_matrix, YN, return_lower_triangular=True)
    # wls calculation, by calculating S^2 from OLS fit, then passing weights
    D_o, _ = ols_fit_tensor(design_matrix, YN, return_lower_triangular=True)
    pred_s = np.exp(np.dot(design_matrix, D_o.T)).T
    D_W, _ = wls_fit_tensor(
        design_matrix, YN, return_lower_triangular=True, weights=pred_s**2
    )  # weights match WLS default
    npt.assert_almost_equal(D_w, D_W)
    # wls calculation, but passing incorrect weights
    D_W, _ = wls_fit_tensor(
        design_matrix, YN, return_lower_triangular=True, weights=pred_s**1
    )
    npt.assert_raises(AssertionError, npt.assert_array_equal, D_w, D_W)
    # Test that wls implementation is correct, by comparison with result here
    W = np.diag(pred_s.squeeze() ** 2)
    AT_W = np.dot(design_matrix.T, W)
    inv_AT_W_A = np.linalg.pinv(np.dot(AT_W, design_matrix))
    AT_W_LS = np.dot(AT_W, np.log(YN).squeeze())
    result = np.dot(inv_AT_W_A, AT_W_LS)
    npt.assert_almost_equal(D_w.squeeze(), result)


def test_rwls_rnlls_irls_fit():
    # Recall: D = [Dxx,Dyy,Dzz,Dxy,Dxz,Dyz,log(S_0)] and D ~ 10^-4 mm^2 /s
    b0 = 1000.0
    bval, bvec = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    B = bval[1]
    # Scale the eigenvalues and tensor by the B value so the units match
    D = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 1.0, -np.log(b0) * B]) / B
    evals = np.array([2.0, 1.0, 0.0]) / B
    md = evals.mean()
    tensor = from_lower_triangular(D)
    # Design Matrix
    gtab = grad.gradient_table(bval, bvecs=bvec)
    X = dti.design_matrix(gtab)
    # Signals
    Y = np.exp(np.dot(X, D))
    npt.assert_almost_equal(Y[0], b0)
    Y.shape = (-1,) + Y.shape

    YN = Y + 1 * np.random.normal(size=Y.shape)  # error, or weights irrelevant
    YN[0, -1] *= 10  # note 1D array!

    for a, ar in zip(["WLS", "NLLS"], ["RWLS", "RNLLS"]):
        # Estimate tensor from test signals
        model = TensorModel(gtab, fit_method=a, return_S0_hat=True)
        tensor_est = model.fit(YN)

        model = TensorModel(gtab, fit_method=ar, return_S0_hat=True, num_iter=10)
        tensor_est_R = model.fit(YN)

        npt.assert_array_less(
            np.linalg.norm(tensor_est_R.evals[0] - evals),
            np.linalg.norm(tensor_est.evals[0] - evals),
        )

        npt.assert_array_less(
            np.linalg.norm(tensor_est_R.quadratic_form[0] - tensor),
            np.linalg.norm(tensor_est.quadratic_form[0] - tensor),
        )

        npt.assert_array_less(
            np.linalg.norm(tensor_est_R.md[0] - md),
            np.linalg.norm(tensor_est.md[0] - md),
        )

        # error is often almost exactly the same, so this test sometimes fails
        # npt.assert_array_less(np.linalg.norm(tensor_est_R.S0_hat[0] - b0),
        #                       np.linalg.norm(tensor_est.S0_hat[0] - b0))

    # test RWLS/RNLLS implemented explicitly via IRLS function
    for wm, fit_type in zip(
        [weights_method_wls_m_est, weights_method_nlls_m_est], ["WLS", "NLLS"]
    ):
        # IRLS implementation
        model = TensorModel(
            gtab,
            fit_method="IRLS",
            return_S0_hat=True,
            weights_method=wm,
            fit_type=fit_type,
            num_iter=10,
        )
        tensor_est_R1 = model.fit(YN)
        npt.assert_equal(tensor_est_R1.model.extra["robust"].shape, YN.shape)
        npt.assert_equal(tensor_est_R1.model.extra["robust"][0, -1], 0)

        # 'shortcut' method RWLS/RNLLS
        model = TensorModel(
            gtab,
            fit_method="R" + fit_type,
            return_S0_hat=False,  # NOTE increase coverage
            num_iter=10,
        )
        tensor_est_R2 = model.fit(YN)
        npt.assert_equal(tensor_est_R2.model.extra["robust"].shape, YN.shape)
        npt.assert_equal(tensor_est_R2.model.extra["robust"][0, -1], 0)

        npt.assert_almost_equal(tensor_est_R1.evals[0], tensor_est_R2.evals[0])

        npt.assert_almost_equal(
            tensor_est_R1.quadratic_form[0], tensor_est_R2.quadratic_form[0]
        )

    # test that error is raised if not enough data
    model = TensorModel(gtab, fit_method="RWLS", num_iter=10)
    npt.assert_raises(ValueError, model.fit, YN[:, 0:3])

    # force use of iter_fit_tensor without making a large test
    model = TensorModel(gtab, fit_method="RWLS", return_S0_hat=True, num_iter=10)
    tensor_est_R2 = model.fit(np.repeat(YN, repeats=1e4 + 1, axis=0))


def test_masked_array_with_tensor():
    data = np.ones((2, 4, 56))
    mask = np.array([[True, False, False, True], [True, False, True, False]])

    bval, bvec = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table_from_bvals_bvecs(bval, bvec)

    # test self.extra with mask by using return_leverages
    tensor_model = TensorModel(gtab, return_leverages=True)
    tensor = tensor_model.fit(data, mask=mask)
    npt.assert_equal(tensor.shape, (2, 4))
    npt.assert_equal(tensor.fa.shape, (2, 4))
    npt.assert_equal(tensor.evals.shape, (2, 4, 3))
    npt.assert_equal(tensor.evecs.shape, (2, 4, 3, 3))

    tensor = tensor[0]
    npt.assert_equal(tensor.shape, (4,))
    npt.assert_equal(tensor.fa.shape, (4,))
    npt.assert_equal(tensor.evals.shape, (4, 3))
    npt.assert_equal(tensor.evecs.shape, (4, 3, 3))

    tensor = tensor[0]
    npt.assert_equal(tensor.shape, ())
    npt.assert_equal(tensor.fa.shape, ())
    npt.assert_equal(tensor.evals.shape, (3,))
    npt.assert_equal(tensor.evecs.shape, (3, 3))
    npt.assert_equal(type(tensor.model_params), np.ndarray)


def test_fit_method_error():
    bval, bvec = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table_from_bvals_bvecs(bval, bvec)

    # This should work (smoke-testing!):
    TensorModel(gtab, fit_method="WLS")

    # This should raise an error because there is no such fit_method
    npt.assert_raises(ValueError, TensorModel, gtab, min_signal=1e-9, fit_method="s")


def test_lower_triangular():
    tensor = np.arange(9).reshape((3, 3))
    D = lower_triangular(tensor)
    npt.assert_array_equal(D, [0, 3, 4, 6, 7, 8])
    D = lower_triangular(tensor, b0=1)
    npt.assert_array_equal(D, [0, 3, 4, 6, 7, 8, 0])
    npt.assert_raises(ValueError, lower_triangular, np.zeros((2, 3)))
    shape = (4, 5, 6)
    many_tensors = np.empty(shape + (3, 3))
    many_tensors[:] = tensor
    result = np.empty(shape + (6,))
    result[:] = [0, 3, 4, 6, 7, 8]
    D = lower_triangular(many_tensors)
    npt.assert_array_equal(D, result)
    D = lower_triangular(many_tensors, b0=1)
    result = np.empty(shape + (7,))
    result[:] = [0, 3, 4, 6, 7, 8, 0]
    npt.assert_array_equal(D, result)


def test_from_lower_triangular():
    result = np.array([[0, 1, 3], [1, 2, 4], [3, 4, 5]])
    D = np.arange(7)
    tensor = from_lower_triangular(D)
    npt.assert_array_equal(tensor, result)
    result = result * np.ones((5, 4, 1, 1))
    D = D * np.ones((5, 4, 1))
    tensor = from_lower_triangular(D)
    npt.assert_array_equal(tensor, result)


def test_all_constant():
    bvals, bvecs = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table_from_bvals_bvecs(bvals, bvecs)
    fit_methods = ["LS", "OLS", "NNLS", "RESTORE"]
    for _ in fit_methods:
        dm = dti.TensorModel(gtab)
        npt.assert_almost_equal(dm.fit(100 * np.ones(bvals.shape[0])).fa, 0)
        # Doesn't matter if the signal is smaller than 1:
        npt.assert_almost_equal(dm.fit(0.4 * np.ones(bvals.shape[0])).fa, 0)


def test_all_zeros():
    bvals, bvecs = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table_from_bvals_bvecs(bvals, bvecs)
    fit_methods = ["LS", "OLS", "NNLS", "RESTORE"]
    for _ in fit_methods:
        dm = dti.TensorModel(gtab)
        evals = dm.fit(np.zeros(bvals.shape[0])).evals
        npt.assert_array_almost_equal(evals, 0)


def test_mask():
    data, gtab = dsi_voxels()
    for fit_type in ["LS", "NLLS"]:
        dm = dti.TensorModel(gtab, fit_method=fit_type)
        mask = np.zeros(data.shape[:-1], dtype=bool)
        mask[0, 0, 0] = True
        dtifit = dm.fit(data)
        dtifit_w_mask = dm.fit(data, mask=mask)
        # Without a mask it has some value
        assert not np.isnan(dtifit.fa[0, 0, 0])
        # Where mask is False, evals, evecs and fa should all be 0
        npt.assert_array_equal(dtifit_w_mask.evals[~mask], 0)
        npt.assert_array_equal(dtifit_w_mask.evecs[~mask], 0)
        npt.assert_array_equal(dtifit_w_mask.fa[~mask], 0)
        # Except for the one voxel that was selected by the mask:
        npt.assert_almost_equal(dtifit_w_mask.fa[0, 0, 0], dtifit.fa[0, 0, 0])

        # Test with returning S0_hat
        dm = dti.TensorModel(gtab, fit_method=fit_type, return_S0_hat=True)
        mask = np.zeros(data.shape[:-1], dtype=bool)
        mask[0, 0, 0] = True
        for mask_more in [True, False]:
            if mask_more:
                mask[0, 0, 1] = True
            dtifit = dm.fit(data)
            dtifit_w_mask = dm.fit(data, mask=mask)
            # Without a mask it has some value
            assert not np.isnan(dtifit.fa[0, 0, 0])
            # Where mask is False, evals, evecs and fa should all be 0
            npt.assert_array_equal(dtifit_w_mask.evals[~mask], 0)
            npt.assert_array_equal(dtifit_w_mask.evecs[~mask], 0)
            npt.assert_array_equal(dtifit_w_mask.fa[~mask], 0)
            npt.assert_array_equal(dtifit_w_mask.S0_hat[~mask], 0)
            # Except for the one voxel that was selected by the mask:
            npt.assert_almost_equal(dtifit_w_mask.fa[0, 0, 0], dtifit.fa[0, 0, 0])
            npt.assert_almost_equal(
                dtifit_w_mask.S0_hat[0, 0, 0], dtifit.S0_hat[0, 0, 0]
            )


@set_random_number_generator()
def test_nnls_jacobian_func(rng):
    b0 = 1000.0
    bval, bvecs = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table(bval, bvecs=bvecs)
    B = bval[1]

    # Scale the eigenvalues and tensor by the B value so the units match
    D_orig = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 1.0, -np.log(b0) * B]) / B

    # Design Matrix
    X = dti.design_matrix(gtab)

    # Signals
    Y = np.exp(np.dot(X, D_orig))
    scale = 10
    error = rng.normal(scale=scale, size=Y.shape)
    Y = Y + error

    nlls = dti._NllsHelper()
    sigma_scalar = 1.4826 * np.median(np.abs(error - np.median(error)))
    sigma_array = np.full_like(Y, sigma_scalar)
    for sigma in [sigma_scalar, sigma_array]:
        weights = 1 / sigma**2
        for D in [D_orig, np.zeros_like(D_orig)]:
            # Test Jacobian at D
            args = [D, X, Y, weights]
            # 1. call 'err_func', to set internal stuff in the class
            nlls.err_func(*args)
            # 2. call 'jabobian_func', corresponds to last err_func call
            # analytical = nlls.jacobian_func(*args)

            # test analytical gradient (needs to be performed per data-point)
            for i in range(len(X)):
                args = [X[i], Y[i], weights]

                # FIXME: this is sometimes failing in tests on Github
                # approx = opt.approx_fprime(D, nlls.err_func, 1e-8, *args)
                #
                #        approx_fprime wants nlls.err_func to return a scalar
                #        value, which it ought to do if called with a single
                #        data point (otherwise, it returns an array, consistent
                #        with scipy.opt.leastsq) but something seems broken in
                #        some tests, so let's make a function that ensures a
                #        scalar is returned. Issue for this *test*, not for
                #        nlls.err_func, which works correctly
                # def ef(x):
                #     tmp = nlls.err_func(x, *args)
                #     return tmp if np.isscalar(tmp) else tmp[0]

                # NOTE: approx_fprime not accurate enough to pass this test
                #       even though it will pass if using autograd code
                #       to ensure a truly accurate derivative of nlls.err_func
                # approx = opt.approx_fprime(D, ef, 1e-8)
                # assert np.allclose(approx, analytical[i])
                assert True


def test_nlls_fit_tensor():
    """
    Test the implementation of NLLS and RESTORE
    """

    b0 = 1000.0
    bvals, bvecs = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table(bvals, bvecs=bvecs)
    B = bvals[1]

    # Scale the eigenvalues and tensor by the B value so the units match
    D = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 1.0, -np.log(b0) * B]) / B
    evals = np.array([2.0, 1.0, 0.0]) / B
    md = evals.mean()
    tensor = from_lower_triangular(D)

    # Design Matrix
    X = dti.design_matrix(gtab)

    # Signals
    Y = np.exp(np.dot(X, D))
    Y.shape = (-1,) + Y.shape

    # Estimate tensor from test signals and compare against expected result
    # using non-linear least squares:
    tensor_model = dti.TensorModel(gtab, fit_method="NLLS")
    tensor_est = tensor_model.fit(Y)
    npt.assert_equal(tensor_est.shape, Y.shape[:-1])
    npt.assert_array_almost_equal(tensor_est.evals[0], evals)
    npt.assert_array_almost_equal(tensor_est.quadratic_form[0], tensor)
    npt.assert_almost_equal(tensor_est.md[0], md)

    # You can also do this without the Jacobian (though it's slower):
    tensor_model = dti.TensorModel(gtab, fit_method="NLLS", jac=False)
    tensor_est = tensor_model.fit(Y)
    npt.assert_equal(tensor_est.shape, Y.shape[:-1])
    npt.assert_array_almost_equal(tensor_est.evals[0], evals)
    npt.assert_array_almost_equal(tensor_est.quadratic_form[0], tensor)
    npt.assert_almost_equal(tensor_est.md[0], md)

    # Using weights:
    weights = 2 * np.ones_like(Y, dtype=np.float32)
    tensor_model = dti.TensorModel(gtab, fit_method="NLLS", weights=weights)
    tensor_est = tensor_model.fit(Y)
    npt.assert_equal(tensor_est.shape, Y.shape[:-1])
    npt.assert_array_almost_equal(tensor_est.evals[0], evals)
    npt.assert_array_almost_equal(tensor_est.quadratic_form[0], tensor)
    npt.assert_almost_equal(tensor_est.md[0], md)

    # Use NLLS with some actual 4D data:
    data, bvals, bvecs = get_fnames(name="small_25")
    gtab = grad.gradient_table(bvals, bvecs=bvecs)
    tm1 = dti.TensorModel(gtab, fit_method="NLLS")
    dd = load_nifti_data(data)
    tf1 = tm1.fit(dd)
    tm2 = dti.TensorModel(gtab)
    tf2 = tm2.fit(dd)

    npt.assert_array_almost_equal(tf1.fa, tf2.fa, decimal=1)

    # Reduce amount of data, to cause NLLS to fail
    gtab_less = grad.gradient_table(gtab.bvals[0:3], bvecs=gtab.bvecs[0:3, :])
    Y_less = Y[..., 0:3].copy()

    # Test warning for failure of NLLS method, resort to OLS result
    # (reason for failure: too few data points for NLLS)
    tensor_model = dti.TensorModel(gtab_less, fit_method="NLLS", return_S0_hat=True)
    tmf = npt.assert_warns(UserWarning, tensor_model.fit, Y_less)

    # Test fail_is_nan=True, failed NLLS method gives NaN
    tensor_model = dti.TensorModel(
        gtab_less, fit_method="NLLS", return_S0_hat=True, fail_is_nan=True
    )
    tmf = npt.assert_warns(UserWarning, tensor_model.fit, Y_less)
    npt.assert_equal(tmf[0].S0_hat, np.nan)


def test_restore():
    """
    Test the implementation of the RESTORE algorithm
    """
    b0 = 1000.0
    bval, bvecs = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table(bval, bvecs=bvecs)
    B = bval[1]

    # Scale the eigenvalues and tensor by the B value so the units match
    D = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 1.0, -np.log(b0) * B]) / B
    evals = np.array([2.0, 1.0, 0.0]) / B
    tensor = from_lower_triangular(D)

    # Design Matrix
    X = dti.design_matrix(gtab)

    # Signals
    Y = np.exp(np.dot(X, D))
    Y = np.vstack([Y[None, :], Y[None, :]])  # two voxels
    for drop_this in range(1, Y.shape[-1]):
        for jac in [True, False]:
            # RESTORE estimates should be robust to dropping
            this_y = Y.copy()
            this_y[:, drop_this] = 1.0
            for sigma in [
                67.0,
                np.array([67.0]),
                np.ones(this_y.shape[-1]) * 67.0,
                np.array([66.0, 67.0]).reshape((-1, 1)),
            ]:
                tensor_model = dti.TensorModel(
                    gtab, fit_method="restore", jac=jac, sigma=sigma
                )

                tensor_est = tensor_model.fit(this_y)
                npt.assert_array_almost_equal(tensor_est.evals[0], evals, decimal=3)
                npt.assert_array_almost_equal(
                    tensor_est.quadratic_form[0], tensor, decimal=3
                )

                # test recording of robust signals
                npt.assert_equal(tensor_est.model.extra["robust"].shape, Y.shape)

    # If sigma is very small, it still needs to work:
    tensor_model = dti.TensorModel(gtab, fit_method="restore", sigma=0.0001)
    tensor_model.fit(Y.copy())

    # If sigma is very small, it still needs to work (it is estimated):
    tensor_model = dti.TensorModel(gtab, fit_method="restore", sigma=None)
    tensor_model.fit(Y.copy() + np.random.normal(size=Y.shape))

    # Test return_S0_hat
    tensor_model = dti.TensorModel(
        gtab, fit_method="restore", sigma=0.0001, return_S0_hat=True
    )
    tmf = tensor_model.fit(Y.copy())
    npt.assert_almost_equal(tmf[0].S0_hat, b0)

    # Test warning for failure of NLLS method, resort to OLS result
    # (reason for failure: too few data points for NLLS, due to negative sigma)
    tensor_model = dti.TensorModel(
        gtab, fit_method="restore", sigma=-1.0, return_S0_hat=True
    )
    tmf = npt.assert_warns(UserWarning, tensor_model.fit, Y.copy())

    # Test fail_is_nan=True, failed NLLS method gives NaN
    tensor_model = dti.TensorModel(
        gtab, fit_method="restore", sigma=-1.0, return_S0_hat=True, fail_is_nan=True
    )
    tmf = npt.assert_warns(UserWarning, tensor_model.fit, Y.copy())
    npt.assert_equal(tmf[0].S0_hat, np.nan)


def test_adc():
    """
    Test the implementation of the calculation of apparent diffusion
    coefficient
    """
    data, gtab = dsi_voxels()
    dm = dti.TensorModel(gtab, fit_method="LS")
    mask = np.zeros(data.shape[:-1], dtype=bool)
    mask[0, 0, 0] = True
    dtifit = dm.fit(data)
    # The ADC in the principal diffusion direction should be equal to the AD in
    # each voxel:

    pdd0 = dtifit.evecs[0, 0, 0, 0]
    sphere_pdd0 = dps.Sphere(x=pdd0[0], y=pdd0[1], z=pdd0[2])
    npt.assert_array_almost_equal(
        dtifit.adc(sphere_pdd0)[0, 0, 0], dtifit.ad[0, 0, 0], decimal=4
    )

    # Test that it works for cases in which the data is 1D
    dtifit = dm.fit(data[0, 0, 0])
    sphere_pdd0 = dps.Sphere(x=pdd0[0], y=pdd0[1], z=pdd0[2])
    npt.assert_array_almost_equal(dtifit.adc(sphere_pdd0), dtifit.ad, decimal=4)


def test_predict():
    """
    Test model prediction API
    """
    psphere = get_sphere(name="symmetric362")
    bvecs = np.concatenate(([[1, 0, 0]], psphere.vertices))
    bvals = np.zeros(len(bvecs)) + 1000
    bvals[0] = 0
    gtab = grad.gradient_table(bvals, bvecs=bvecs)
    mevals = np.array(([0.0015, 0.0003, 0.0001], [0.0015, 0.0003, 0.0003]))
    mevecs = [
        np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        np.array([[0, 0, 1], [0, 1, 0], [1, 0, 0]]),
    ]
    S = single_tensor(gtab, 100, evals=mevals[0], evecs=mevecs[0], snr=None)

    dm = dti.TensorModel(gtab, fit_method="LS", return_S0_hat=True)
    dmfit = dm.fit(S)
    npt.assert_array_almost_equal(dmfit.predict(gtab, S0=100), S)
    npt.assert_array_almost_equal(dmfit.predict(gtab), S)
    npt.assert_array_almost_equal(dm.predict(dmfit.model_params, S0=100), S)

    fdata, fbvals, fbvecs = get_fnames()
    data = load_nifti_data(fdata)
    # Make the data cube a bit larger:
    data = np.tile(data.T, 2).T
    gtab = grad.gradient_table(fbvals, bvecs=fbvecs)
    dtim = dti.TensorModel(gtab)
    dtif = dtim.fit(data)
    S0 = np.mean(data[..., gtab.b0s_mask], -1)
    p = dtif.predict(gtab, S0=S0)
    npt.assert_equal(p.shape, data.shape)
    # Predict using S0_hat:
    dtim = dti.TensorModel(gtab, return_S0_hat=True)
    dtif = dtim.fit(data)
    p = dtif.predict(gtab)
    npt.assert_equal(p.shape, data.shape)
    p = dtif.predict(gtab, S0=S0)
    npt.assert_equal(p.shape, data.shape)

    # Test iter_fit_tensor with S0_hat
    dtim = dti.TensorModel(gtab, step=2, return_S0_hat=True)
    dtif = dtim.fit(data)
    S0 = np.mean(data[..., gtab.b0s_mask], -1)
    p = dtif.predict(gtab, S0=S0)
    npt.assert_equal(p.shape, data.shape)

    # Use a smaller step in predicting:

    dtim = dti.TensorModel(gtab, step=2)
    dtif = dtim.fit(data)
    S0 = np.mean(data[..., gtab.b0s_mask], -1)
    p = dtif.predict(gtab, S0=S0)
    npt.assert_equal(p.shape, data.shape)
    # And with a scalar S0:
    S0 = 1
    p = dtif.predict(gtab, S0=S0)
    npt.assert_equal(p.shape, data.shape)
    # Assign the step through kwarg:
    p = dtif.predict(gtab, S0=S0, step=1)
    npt.assert_equal(p.shape, data.shape)
    # And without S0:
    p = dtif.predict(gtab, step=1)
    npt.assert_equal(p.shape, data.shape)


def test_eig_from_lo_tri():
    psphere = get_sphere(name="symmetric362")
    bvecs = np.concatenate(([[0, 0, 0]], psphere.vertices))
    bvals = np.zeros(len(bvecs)) + 1000
    bvals[0] = 0
    gtab = grad.gradient_table(bvals, bvecs=bvecs)
    mevals = np.array(([0.0015, 0.0003, 0.0001], [0.0015, 0.0003, 0.0003]))
    mevecs = [
        np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]),
        np.array([[0, 0, 1], [0, 1, 0], [1, 0, 0]]),
    ]
    S = np.array(
        [
            [
                single_tensor(gtab, 100, evals=mevals[0], evecs=mevecs[0], snr=None),
                single_tensor(gtab, 100, evals=mevals[0], evecs=mevecs[0], snr=None),
            ]
        ]
    )

    dm = dti.TensorModel(gtab, fit_method="LS")
    dmfit = dm.fit(S)

    lo_tri = lower_triangular(dmfit.quadratic_form)
    npt.assert_array_almost_equal(dti.eig_from_lo_tri(lo_tri), dmfit.model_params)


def test_min_signal_alone():
    fdata, fbvals, fbvecs = get_fnames()
    data = load_nifti_data(fdata)
    gtab = grad.gradient_table(fbvals, bvecs=fbvecs)

    idx = tuple(np.array(np.where(data == np.min(data)))[:-1, 0])
    ten_model = dti.TensorModel(gtab)
    fit_alone = ten_model.fit(data[idx])
    fit_together = ten_model.fit(data)
    npt.assert_array_almost_equal(
        fit_together.model_params[idx], fit_alone.model_params, decimal=12
    )


def test_decompose_tensor_nan():
    D_fine = np.array([1.7e-3, 0.0, 0.3e-3, 0.0, 0.0, 0.2e-3])
    D_alter = np.array([1.6e-3, 0.0, 0.4e-3, 0.0, 0.0, 0.3e-3])
    D_nan = np.nan * np.ones(6)

    lref, vref = decompose_tensor(from_lower_triangular(D_fine))
    lfine, vfine = _decompose_tensor_nan(
        from_lower_triangular(D_fine), from_lower_triangular(D_alter)
    )
    npt.assert_array_almost_equal(lfine, np.array([1.7e-3, 0.3e-3, 0.2e-3]))
    npt.assert_array_almost_equal(vfine, vref)

    lref, vref = decompose_tensor(from_lower_triangular(D_alter))
    lalter, valter = _decompose_tensor_nan(
        from_lower_triangular(D_nan), from_lower_triangular(D_alter)
    )
    npt.assert_array_almost_equal(lalter, np.array([1.6e-3, 0.4e-3, 0.3e-3]))
    npt.assert_array_almost_equal(valter, vref)


def test_design_matrix_lte():
    _, fbval, fbvec = get_fnames(name="small_25")
    gtab_btens_none = grad.gradient_table(fbval, bvecs=fbvec)
    gtab_btens_lte = grad.gradient_table(fbval, bvecs=fbvec, btens="LTE")

    B_btens_none = dti.design_matrix(gtab_btens_none)
    B_btens_lte = dti.design_matrix(gtab_btens_lte)
    npt.assert_array_almost_equal(B_btens_none, B_btens_lte, decimal=1)


def test_extra_return():
    """
    Test if returns of dictionary 'extra' from fitting functions are working
    properly.

    Uses data/55dir_grad as the gradient table and 3by3by56.nii
    as the data.

    """

    b0 = 1000.0
    bval, bvecs = read_bvals_bvecs(*get_fnames(name="55dir_grad"))
    gtab = grad.gradient_table(bval, bvecs=bvecs)
    B = bval[1]

    # Scale the eigenvalues and tensor by the B value so the units match
    D = np.array([1.0, 1.0, 1.0, 0.0, 0.0, 1.0, -np.log(b0) * B]) / B

    # Design Matrix
    X = dti.design_matrix(gtab)

    # Signals
    Y = np.exp(np.dot(X, D))
    Y = np.vstack([Y[None, :], Y[None, :]])  # two voxels
    for drop_this in range(1, 3):  # Y.shape[-1]):
        # test specific extra from specific methods
        for method in ["restore"]:
            this_y = Y.copy()
            this_y[:, drop_this] = 1.0

            sigma = 0.0001

            if method == "restore":
                tensor_model = dti.TensorModel(gtab, fit_method=method, sigma=sigma)

            tensor_est = tensor_model.fit(this_y)
            npt.assert_equal(tensor_est.model.extra["robust"].shape, Y.shape)



from numpy.testing import assert_equal

from dipy.utils.compatibility import check_max_version, check_min_version


def test_check_min_version():
    assert_equal(check_min_version("dipy", "15.0.0"), False)
    assert_equal(check_min_version("dipy", "0.8.0"), True)

    assert_equal(check_min_version("numpy", "15.0.0"), False)
    assert_equal(check_min_version("numpy", "1.8.0"), True)

    assert_equal(check_min_version("scipy", "3.0.0"), False)
    assert_equal(check_min_version("scipy", "1.8.0"), True)

    assert_equal(check_min_version("fakepackage", "3.0.0"), False)
    assert_equal(check_min_version("fakepackage", "0.8.0"), False)


def test_check_max_version():
    assert_equal(check_max_version("dipy", "15.0.0"), True)
    assert_equal(check_max_version("dipy", "0.8.0"), False)

    assert_equal(check_max_version("numpy", "15.0.0"), True)
    assert_equal(check_max_version("numpy", "1.8.0"), False)

    assert_equal(check_max_version("scipy", "3.0.0"), True)
    assert_equal(check_max_version("scipy", "1.8.0"), False)

    assert_equal(check_max_version("fakepackage", "3.0.0"), False)
    assert_equal(check_max_version("fakepackage", "0.8.0"), False)



import numpy as np

floating = np.float32


class Bunch:
    def __init__(self, **kwds):
        r"""A 'bunch' of values (a replacement of Enum)

        This is a temporary replacement of Enum, which is not available
        on all versions of Python 2
        """
        self.__dict__.update(kwds)


VerbosityLevels = Bunch(NONE=0, STATUS=1, DIAGNOSE=2, DEBUG=3)
r""" VerbosityLevels
This enum defines the four levels of verbosity we use in the align
module.
NONE : do not print anything
STATUS : print information about the current status of the algorithm
DIAGNOSE : print high level information of the components involved in the
registration that can be used to detect a failing component.
DEBUG : print as much information as possible to isolate the cause of a bug.
"""

from dipy.align._public import (
    affine,
    affine_registration,
    center_of_mass,
    motion_correction,
    read_mapping,
    register_dwi_series,
    register_dwi_to_template,  # noqa
    register_series,
    resample,
    rigid,
    rigid_isoscaling,
    rigid_scaling,
    streamline_registration,
    syn_registration,
    translation,
    write_mapping,
)

__all__ = [
    "syn_registration",
    "register_dwi_to_template",
    "write_mapping",
    "read_mapping",
    "resample",
    "center_of_mass",
    "translation",
    "rigid_isoscaling",
    "rigid_scaling",
    "rigid",
    "affine",
    "motion_correction",
    "affine_registration",
    "register_series",
    "register_dwi_series",
    "streamline_registration",
]



import logging
from os.path import join as pjoin
from warnings import warn

import nibabel as nib
import numpy as np

from dipy.align import affine_registration, motion_correction
from dipy.align.imaffine import AffineMap
from dipy.align.imwarp import DiffeomorphicMap, SymmetricDiffeomorphicRegistration
from dipy.align.metrics import CCMetric, EMMetric, SSDMetric
from dipy.align.reslice import reslice
from dipy.align.streamlinear import slr_with_qbx
from dipy.align.streamwarp import bundlewarp
from dipy.core.gradients import gradient_table, mask_non_weighted_bvals
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti, save_qa_metric
from dipy.tracking.streamline import set_number_of_points, transform_streamlines
from dipy.utils.optpkg import optional_package
from dipy.workflows.utils import handle_vol_idx
from dipy.workflows.workflow import Workflow

pd, have_pd, _ = optional_package("pandas")


def check_dimensions(static, moving):
    """Check the dimensions of the input images.

    Parameters
    ----------
    static : 2D or 3D array
        the image to be used as reference during optimization.

    moving: 2D or 3D array
        the image to be used as "moving" during optimization. It is
        necessary to pre-align the moving image to ensure its domain
        lies inside the domain of the deformation fields. This is assumed
        to be accomplished by "pre-aligning" the moving image towards the
        static using an affine transformation given by the
        'starting_affine' matrix.

    """
    if len(static.shape) != len(moving.shape):
        raise ValueError(
            "Dimension mismatch: The input images must "
            "have same number of dimensions."
        )

    if len(static.shape) > 3 and len(moving.shape) > 3:
        raise ValueError(
            "Dimension mismatch: One of the input should be 2D or 3D dimensions."
        )


class ResliceFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "reslice"

    def run(
        self,
        input_files,
        new_vox_size,
        order=1,
        mode="constant",
        cval=0,
        num_processes=1,
        out_dir="",
        out_resliced="resliced.nii.gz",
    ):
        """Reslice data with new voxel resolution defined by ``new_vox_sz``

        Parameters
        ----------
        input_files : string
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        new_vox_size : variable float
            new voxel size.
        order : int, optional
            order of interpolation, from 0 to 5, for resampling/reslicing,
            0 nearest interpolation, 1 trilinear etc.. if you don't want any
            smoothing 0 is the option you need.
        mode : string, optional
            Points outside the boundaries of the input are filled according
            to the given mode 'constant', 'nearest', 'reflect' or 'wrap'.
        cval : float, optional
            Value used for points outside the boundaries of the input if
            mode='constant'.
        num_processes : int, optional
            Split the calculation to a pool of children processes. This only
            applies to 4D `data` arrays. Default is 1. If < 0 the maximal
            number of cores minus ``num_processes + 1`` is used (enter -1 to
            use as many cores as possible). 0 raises an error.
        out_dir : string, optional
            Output directory.
        out_resliced : string, optional
            Name of the resliced dataset to be saved.
        """

        io_it = self.get_io_iterator()

        for inputfile, outpfile in io_it:
            data, affine, vox_sz = load_nifti(inputfile, return_voxsize=True)
            logging.info(f"Processing {inputfile}")
            new_data, new_affine = reslice(
                data,
                affine,
                vox_sz,
                new_vox_size,
                order=order,
                mode=mode,
                cval=cval,
                num_processes=num_processes,
            )
            save_nifti(outpfile, new_data, new_affine)
            logging.info(f"Resliced file save in {outpfile}")


class SlrWithQbxFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "slrwithqbx"

    def run(
        self,
        static_files,
        moving_files,
        x0="affine",
        rm_small_clusters=50,
        qbx_thr=(40, 30, 20, 15),
        num_threads=None,
        greater_than=50,
        less_than=250,
        nb_pts=20,
        progressive=True,
        out_dir="",
        out_moved="moved.trk",
        out_affine="affine.txt",
        out_stat_centroids="static_centroids.trk",
        out_moving_centroids="moving_centroids.trk",
        out_moved_centroids="moved_centroids.trk",
    ):
        """Streamline-based linear registration.

        For efficiency we apply the registration on cluster centroids and
        remove small clusters.

        See :footcite:p:`Garyfallidis2014b`, :footcite:p:`Garyfallidis2015`,
        :footcite:p:`Garyfallidis2018` for further details.

        Parameters
        ----------
        static_files : string
            List of reference/fixed bundle tractograms.
        moving_files : string
            List of target bundle tractograms that will be moved/registered to
            match the static bundles.
        x0 : string, optional
            rigid, similarity or affine transformation model.
        rm_small_clusters : int, optional
            Remove clusters that have less than `rm_small_clusters`.
        qbx_thr : variable int, optional
            Thresholds for QuickBundlesX.
        num_threads : int, optional
            Number of threads to be used for OpenMP parallelization. If None
            (default) the value of OMP_NUM_THREADS environment variable is
            used if it is set, otherwise all available threads are used. If
            < 0 the maximal number of threads minus $|num_threads + 1|$ is used
            (enter -1 to use as many threads as possible). 0 raises an error.
            Only metrics using OpenMP will use this variable.
        greater_than : int, optional
            Keep streamlines that have length greater than
            this value.
        less_than : int, optional
            Keep streamlines have length less than this value.
        nb_pts : int, optional
            Number of points for discretizing each streamline.
        progressive : boolean, optional
            True to enable progressive registration.
        out_dir : string, optional
            Output directory.
        out_moved : string, optional
            Filename of moved tractogram.
        out_affine : string, optional
            Filename of affine for SLR transformation.
        out_stat_centroids : string, optional
            Filename of static centroids.
        out_moving_centroids : string, optional
            Filename of moving centroids.
        out_moved_centroids : string, optional
            Filename of moved centroids.

        Notes
        -----
        The order of operations is the following. First short or long
        streamlines are removed. Second the tractogram or a random selection
        of the tractogram is clustered with QuickBundlesX. Then SLR
        :footcite:p:`Garyfallidis2015` is applied.

        References
        ----------
        .. footbibliography::
        """

        io_it = self.get_io_iterator()

        logging.info("QuickBundlesX clustering is in use")
        logging.info(f"QBX thresholds {qbx_thr}")

        for (
            static_file,
            moving_file,
            out_moved_file,
            out_affine_file,
            static_centroids_file,
            moving_centroids_file,
            moved_centroids_file,
        ) in io_it:
            logging.info(f"Loading static file {static_file}")
            logging.info(f"Loading moving file {moving_file}")

            static_obj = nib.streamlines.load(static_file)
            moving_obj = nib.streamlines.load(moving_file)

            static, static_header = static_obj.streamlines, static_obj.header
            moving, moving_header = moving_obj.streamlines, moving_obj.header

            moved, affine, centroids_static, centroids_moving = slr_with_qbx(
                static,
                moving,
                x0=x0,
                rm_small_clusters=rm_small_clusters,
                greater_than=greater_than,
                less_than=less_than,
                qbx_thr=qbx_thr,
            )

            logging.info(f"Saving output file {out_moved_file}")
            new_tractogram = nib.streamlines.Tractogram(
                moved, affine_to_rasmm=np.eye(4)
            )
            nib.streamlines.save(new_tractogram, out_moved_file, header=moving_header)

            logging.info(f"Saving output file {out_affine_file}")
            np.savetxt(out_affine_file, affine)

            logging.info(f"Saving output file {static_centroids_file}")
            new_tractogram = nib.streamlines.Tractogram(
                centroids_static, affine_to_rasmm=np.eye(4)
            )
            nib.streamlines.save(
                new_tractogram, static_centroids_file, header=static_header
            )

            logging.info(f"Saving output file {moving_centroids_file}")
            new_tractogram = nib.streamlines.Tractogram(
                centroids_moving, affine_to_rasmm=np.eye(4)
            )
            nib.streamlines.save(
                new_tractogram, moving_centroids_file, header=moving_header
            )

            centroids_moved = transform_streamlines(centroids_moving, affine)

            logging.info(f"Saving output file {moved_centroids_file}")

            new_tractogram = nib.streamlines.Tractogram(
                centroids_moved, affine_to_rasmm=np.eye(4)
            )
            nib.streamlines.save(
                new_tractogram, moved_centroids_file, header=moving_header
            )


class ImageRegistrationFlow(Workflow):
    """
    The registration workflow allows the user to use only one type of
    registration (such as center of mass or rigid body registration only).

    Alternatively, a registration can be done in a progressive manner.
    For example, using affine registration with progressive set to 'True'
    will involve center of mass, translation, rigid body and full affine
    registration. Whereas, when progressive is False the registration will
    include only center of mass and affine registration. The progressive
    registration will be slower but will improve the quality.

    This can be controlled by using the progressive flag (True by default).
    """

    def run(
        self,
        static_image_files,
        moving_image_files,
        transform="affine",
        nbins=32,
        sampling_prop=None,
        metric="mi",
        level_iters=(10000, 1000, 100),
        sigmas=(3.0, 1.0, 0.0),
        factors=(4, 2, 1),
        progressive=True,
        save_metric=False,
        static_vol_idx=None,
        moving_vol_idx=None,
        out_dir="",
        out_moved="moved.nii.gz",
        out_affine="affine.txt",
        out_quality="quality_metric.txt",
    ):
        """
        Parameters
        ----------
        static_image_files : string
            Path to the static image file.
        moving_image_files : string
            Path to the moving image file.
        transform : string, optional
            ``'com'``: center of mass; ``'trans'``: translation; ``'rigid'``:
            rigid body; ``'rigid_isoscaling'``: rigid body + isotropic scaling,
            ``'rigid_scaling'``: rigid body + scaling; ``'affine'``: full affine
            including translation, rotation, shearing and scaling.
        nbins : int, optional
            Number of bins to discretize the joint and marginal PDF.
        sampling_prop : int, optional
            Number ([0-100]) of voxels for calculating the PDF. None implies all
            voxels.
        metric : string, optional
            Similarity metric for gathering mutual information.
        level_iters : variable int, optional
            The number of iterations at each scale of the scale space.
            `level_iters[0]` corresponds to the coarsest scale,
            `level_iters[-1]` the finest, where n is the length of the
            sequence.
        sigmas : variable floats, optional
            Custom smoothing parameter to build the scale space (one parameter
            for each scale).
        factors : variable floats, optional
            Custom scale factors to build the scale space (one factor for each scale).
        progressive : boolean, optional
            Enable/Disable the progressive registration.
        save_metric : boolean, optional
            If true, quality assessment metric are saved in
            'quality_metric.txt'.
        static_vol_idx : str, optional
            1D array representing indices of ``axis=-1`` of a 4D
            `static` input volume. From the command line use something like
            `3 4 5 6`. From script use something like `[3, 4, 5, 6]`. This
            input is required for 4D volumes.
        moving_vol_idx : str, optional
            1D array representing indices of ``axis=-1`` of a 4D
            `moving` input volume. From the command line use something like
            `3 4 5 6`. From script use something like `[3, 4, 5, 6]`. This
            input is required for 4D volumes.
        out_dir : string, optional
            Directory to save the transformed image and the affine matrix
        out_moved : string, optional
            Name for the saved transformed image.
        out_affine : string, optional
            Name for the saved affine matrix.
        out_quality : string, optional
            Name of the file containing the saved quality metric.
        """

        io_it = self.get_io_iterator()
        transform = transform.lower()
        metric = metric.upper()
        if metric != "MI":
            raise ValueError(
                "Invalid similarity metric: Please provide a" "valid metric."
            )

        if progressive:
            pipeline_opt = {
                "com": ["center_of_mass"],
                "trans": ["center_of_mass", "translation"],
                "rigid": ["center_of_mass", "translation", "rigid"],
                "rigid_isoscaling": [
                    "center_of_mass",
                    "translation",
                    "rigid_isoscaling",
                ],
                "rigid_scaling": ["center_of_mass", "translation", "rigid_scaling"],
                "affine": ["center_of_mass", "translation", "rigid", "affine"],
            }
        else:
            pipeline_opt = {
                "com": ["center_of_mass"],
                "trans": ["center_of_mass", "translation"],
                "rigid": ["center_of_mass", "rigid"],
                "rigid_isoscaling": ["center_of_mass", "rigid_isoscaling"],
                "rigid_scaling": ["center_of_mass", "rigid_scaling"],
                "affine": ["center_of_mass", "affine"],
            }

        static_vol_idx = handle_vol_idx(static_vol_idx)

        moving_vol_idx = handle_vol_idx(moving_vol_idx)

        pipeline = pipeline_opt.get(transform)

        if pipeline is None:
            raise ValueError(
                "Invalid transformation:"
                " Please see program's help"
                " for allowed values of"
                " transformation."
            )

        for static_img, mov_img, moved_file, affine_matrix_file, qual_val_file in io_it:
            # Load the data from the input files and store into objects.
            static, static_grid2world = load_nifti(static_img)
            moving, moving_grid2world = load_nifti(mov_img)

            if static_vol_idx is not None:
                static = static[..., static_vol_idx].mean(axis=-1)
            if moving_vol_idx is not None:
                moving = moving[..., moving_vol_idx].mean(axis=-1)

            check_dimensions(static, moving)

            starting_affine = None

            # If only center of mass is selected do not return metric
            if pipeline == ["center_of_mass"]:
                moved_image, affine_matrix = affine_registration(
                    moving,
                    static,
                    moving_affine=moving_grid2world,
                    static_affine=static_grid2world,
                    pipeline=pipeline,
                    starting_affine=starting_affine,
                    metric=metric,
                    level_iters=level_iters,
                    sigmas=sigmas,
                    factors=factors,
                    nbins=nbins,
                    sampling_proportion=sampling_prop,
                )
            else:
                moved_image, affine_matrix, xopt, fopt = affine_registration(
                    moving,
                    static,
                    moving_affine=moving_grid2world,
                    static_affine=static_grid2world,
                    pipeline=pipeline,
                    starting_affine=starting_affine,
                    metric=metric,
                    level_iters=level_iters,
                    sigmas=sigmas,
                    factors=factors,
                    ret_metric=True,
                    nbins=nbins,
                    sampling_proportion=sampling_prop,
                )

                """
                Saving the moved image file and the affine matrix.
                """
                logging.info(f"Optimal parameters: {str(xopt)}")
                logging.info(f"Similarity metric: {str(fopt)}")

                if save_metric:
                    save_qa_metric(qual_val_file, xopt, fopt)

            save_nifti(moved_file, moved_image, static_grid2world)
            np.savetxt(affine_matrix_file, affine_matrix)


class ApplyTransformFlow(Workflow):
    def run(
        self,
        static_image_files,
        moving_image_files,
        transform_map_file,
        transform_type="affine",
        out_dir="",
        out_file="transformed.nii.gz",
    ):
        """
        Parameters
        ----------
        static_image_files : string
            Path of the static image file.
        moving_image_files : string
            Path of the moving image(s). It can be a single image or a
            folder containing multiple images.
        transform_map_file : string
            For the affine case, it should be a text(``*.txt``) file containing
            the affine matrix. For the diffeomorphic case,
            it should be a nifti file containing the mapping displacement
            field in each voxel with this shape (x, y, z, 3, 2).
        transform_type : string, optional
            Select the transformation type to apply between 'affine' or
            'diffeomorphic'.
        out_dir : string, optional
            Directory to save the transformed files.
        out_file : string, optional
            Name of the transformed file.
            It is recommended to use the flag --mix-names to
            prevent the output files from being overwritten.

        """
        if transform_type.lower() not in ["affine", "diffeomorphic"]:
            raise ValueError(
                "Invalid transformation type: Please"
                " provide a valid transform like 'affine'"
                " or 'diffeomorphic'"
            )

        io = self.get_io_iterator()

        for static_image_file, moving_image_file, transform_file, out_file in io:
            # Loading the image data from the input files into object.
            static_image, static_grid2world = load_nifti(static_image_file)
            moving_image, moving_grid2world = load_nifti(moving_image_file)

            # Doing a sanity check for validating the dimensions of the input
            # images.
            if static_image.ndim > moving_image.ndim:
                static_image = static_image[..., 0]
            if static_image.ndim < moving_image.ndim:
                moving_image_full = moving_image
                moving_image = moving_image[..., 0]
            else:
                moving_image_full = None
            check_dimensions(static_image, moving_image)

            if transform_type.lower() == "affine":
                # Loading the affine matrix.
                affine_matrix = np.loadtxt(transform_file)

                # Setting up the affine transformation object.
                mapping = AffineMap(
                    affine=affine_matrix,
                    domain_grid_shape=static_image.shape,
                    domain_grid2world=static_grid2world,
                    codomain_grid_shape=moving_image.shape,
                    codomain_grid2world=moving_grid2world,
                )

            elif transform_type.lower() == "diffeomorphic":
                # Loading the diffeomorphic map.
                disp_data, disp_affine = load_nifti(transform_file)

                mapping = DiffeomorphicMap(
                    3,
                    disp_data.shape[:3],
                    disp_grid2world=np.linalg.inv(disp_affine),
                    domain_shape=static_image.shape,
                    domain_grid2world=static_grid2world,
                    codomain_shape=moving_image.shape,
                    codomain_grid2world=moving_grid2world,
                )

                mapping.forward = disp_data[..., 0]
                mapping.backward = disp_data[..., 1]
                mapping.is_inverse = True

            # Transforming the image
            if moving_image_full is None:
                transformed = mapping.transform(moving_image)
            else:
                transformed = np.concatenate(
                    [
                        mapping.transform(moving_image)[..., None]
                        for moving_image in moving_image_full
                    ],
                    axis=-1,
                )

            save_nifti(out_file, transformed, affine=static_grid2world)


class SynRegistrationFlow(Workflow):
    def run(
        self,
        static_image_files,
        moving_image_files,
        prealign_file="",
        inv_static=False,
        level_iters=(10, 10, 5),
        metric="cc",
        mopt_sigma_diff=2.0,
        mopt_radius=4,
        mopt_smooth=0.0,
        mopt_inner_iter=0,
        mopt_q_levels=256,
        mopt_double_gradient=True,
        mopt_step_type="",
        step_length=0.25,
        ss_sigma_factor=0.2,
        opt_tol=1e-5,
        inv_iter=20,
        inv_tol=1e-3,
        out_dir="",
        out_warped="warped_moved.nii.gz",
        out_inv_static="inc_static.nii.gz",
        out_field="displacement_field.nii.gz",
    ):
        """
        Parameters
        ----------
        static_image_files : string
            Path of the static image file.
        moving_image_files : string
            Path to the moving image file.
        prealign_file : string, optional
            The text file containing pre alignment information via an affine matrix.
        inv_static : boolean, optional
            Apply the inverse mapping to the static image.
        level_iters : variable int, optional
            The number of iterations at each level of the gaussian pyramid.
        metric : string, optional
            The metric to be used.
            metric available: cc (Cross Correlation), ssd (Sum Squared
            Difference), em (Expectation-Maximization).
        mopt_sigma_diff : float, optional
            Metric option applied on Cross correlation (CC).
            The standard deviation of the Gaussian smoothing kernel to be
            applied to the update field at each iteration.
        mopt_radius : int, optional
            Metric option applied on Cross correlation (CC).
            the radius of the squared (cubic) neighborhood at each voxel to
            be considered to compute the cross correlation.
        mopt_smooth : float, optional
            Metric option applied on Sum Squared Difference (SSD) and
            Expectation Maximization (EM). Smoothness parameter, the
            larger the value the smoother the deformation field.
            (default 1.0 for EM, 4.0 for SSD)
        mopt_inner_iter : int, optional
            Metric option applied on Sum Squared Difference (SSD) and
            Expectation Maximization (EM). This is number of iterations to be
            performed at each level of the multi-resolution Gauss-Seidel
            optimization algorithm (this is not the number of steps per
            Gaussian Pyramid level, that parameter must be set for the
            optimizer, not the metric). Default 5 for EM, 10 for SSD.
        mopt_q_levels : int, optional
            Metric option applied on Expectation Maximization (EM).
            Number of quantization levels (Default: 256 for EM)
        mopt_double_gradient : bool, optional
            Metric option applied on Expectation Maximization (EM).
            if True, the gradient of the expected static image under the moving
            modality will be added to the gradient of the moving image,
            similarly, the gradient of the expected moving image under the
            static modality will be added to the gradient of the static image.
        mopt_step_type : string, optional
            Metric option applied on Sum Squared Difference (SSD) and
            Expectation Maximization (EM). The optimization schedule to be
            used in the multi-resolution Gauss-Seidel optimization algorithm
            (not used if Demons Step is selected). Possible value:
            ('gauss_newton', 'demons'). default: 'gauss_newton' for EM,
            'demons' for SSD.
        step_length : float, optional
            the length of the maximum displacement vector of the update
            displacement field at each iteration.
        ss_sigma_factor : float, optional
            parameter of the scale-space smoothing kernel. For example, the
            std. dev. of the kernel will be factor*(2^i) in the isotropic case
            where i = 0, 1, ..., n_scales is the scale.
        opt_tol : float, optional
            the optimization will stop when the estimated derivative of the
            energy profile w.r.t. time falls below this threshold.
        inv_iter : int, optional
            the number of iterations to be performed by the displacement field
            inversion algorithm.
        inv_tol : float, optional
            the displacement field inversion algorithm will stop iterating
            when the inversion error falls below this threshold.
        out_dir : string, optional
            Directory to save the transformed files.
        out_warped : string, optional
            Name of the warped file.
        out_inv_static : string, optional
            Name of the file to save the static image after applying the
            inverse mapping.
        out_field : string, optional
            Name of the file to save the diffeomorphic map.

        """
        io_it = self.get_io_iterator()
        metric = metric.lower()
        if metric not in ["ssd", "cc", "em"]:
            raise ValueError(
                "Invalid similarity metric: Please"
                " provide a valid metric like 'ssd', 'cc', 'em'"
            )

        logging.info("Starting Diffeomorphic Registration")
        logging.info(f"Using {metric.upper()} Metric")

        # Init parameter if they are not setup
        init_param = {
            "ssd": {
                "mopt_smooth": 4.0,
                "mopt_inner_iter": 10,
                "mopt_step_type": "demons",
            },
            "em": {
                "mopt_smooth": 1.0,
                "mopt_inner_iter": 5,
                "mopt_step_type": "gauss_newton",
            },
        }

        mopt_smooth = (
            mopt_smooth
            if mopt_smooth or metric == "cc"
            else init_param[metric]["mopt_smooth"]
        )
        mopt_inner_iter = (
            mopt_inner_iter
            if mopt_inner_iter or metric == "cc"
            else init_param[metric]["mopt_inner_iter"]
        )

        # If using the 'cc' metric, force the `mopt_step_type` parameter to an
        # empty value since the 'cc' metric does not use it; for the rest of
        # the metrics, the `step_type` parameter will be initialized to their
        # corresponding default values in `init_param`.
        if metric == "cc":
            mopt_step_type = ""

        for (
            static_file,
            moving_file,
            owarped_file,
            oinv_static_file,
            omap_file,
        ) in io_it:
            logging.info(f"Loading static file {static_file}")
            logging.info(f"Loading moving file {moving_file}")

            # Loading the image data from the input files into object.
            static_image, static_grid2world = load_nifti(static_file)
            moving_image, moving_grid2world = load_nifti(moving_file)

            # Sanity check for the input image dimensions.
            check_dimensions(static_image, moving_image)

            # Loading the affine matrix.
            prealign = np.loadtxt(prealign_file) if prealign_file else None

            # Note that `step_type` is initialized to the default value in
            # `init_param` for the metric that was not specified as a
            # parameter or if the `mopt_step_type` is empty.
            l_metric = {
                "ssd": SSDMetric(
                    static_image.ndim,
                    smooth=mopt_smooth,
                    inner_iter=mopt_inner_iter,
                    step_type=mopt_step_type
                    if (mopt_step_type and mopt_step_type.strip()) and metric == "ssd"
                    else init_param["ssd"]["mopt_step_type"],
                ),
                "cc": CCMetric(
                    static_image.ndim, sigma_diff=mopt_sigma_diff, radius=mopt_radius
                ),
                "em": EMMetric(
                    static_image.ndim,
                    smooth=mopt_smooth,
                    inner_iter=mopt_inner_iter,
                    step_type=mopt_step_type
                    if (mopt_step_type and mopt_step_type.strip()) and metric == "em"
                    else init_param["em"]["mopt_step_type"],
                    q_levels=mopt_q_levels,
                    double_gradient=mopt_double_gradient,
                ),
            }

            current_metric = l_metric.get(metric.lower())

            sdr = SymmetricDiffeomorphicRegistration(
                metric=current_metric,
                level_iters=level_iters,
                step_length=step_length,
                ss_sigma_factor=ss_sigma_factor,
                opt_tol=opt_tol,
                inv_iter=inv_iter,
                inv_tol=inv_tol,
            )

            mapping = sdr.optimize(
                static_image,
                moving_image,
                static_grid2world=static_grid2world,
                moving_grid2world=moving_grid2world,
                prealign=prealign,
            )

            mapping_data = np.array([mapping.forward.T, mapping.backward.T]).T
            warped_moving = mapping.transform(moving_image)
            mapping.is_inverse = True
            inv_static = mapping.transform(static_image)

            # Saving
            logging.info(f"Saving warped {owarped_file}")
            save_nifti(owarped_file, warped_moving, static_grid2world)
            logging.info(f"Saving inverse transformes static {oinv_static_file}")
            save_nifti(oinv_static_file, inv_static, static_grid2world)
            logging.info(f"Saving Diffeomorphic map {omap_file}")
            save_nifti(omap_file, mapping_data, mapping.codomain_world2grid)


class MotionCorrectionFlow(Workflow):
    """
    The Motion Correction workflow allows the user to align between-volumes
    DWI dataset.
    """

    def run(
        self,
        input_files,
        bvalues_files,
        bvectors_files,
        b0_threshold=50,
        bvecs_tol=0.01,
        out_dir="",
        out_moved="moved.nii.gz",
        out_affine="affine.txt",
    ):
        """
        Parameters
        ----------
        input_files : string
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        bvalues_files : string
            Path to the bvalues files. This path may contain wildcards to use
            multiple bvalues files at once.
        bvectors_files : string
            Path to the bvectors files. This path may contain wildcards to use
            multiple bvectors files at once.
        b0_threshold : float, optional
            Threshold used to find b0 volumes.
        bvecs_tol : float, optional
            Threshold used to check that norm(bvec) = 1 +/- bvecs_tol
            b-vectors are unit vectors
        out_dir : string, optional
            Directory to save the transformed image and the affine matrix.
        out_moved : string, optional
            Name for the saved transformed image.
        out_affine : string, optional
            Name for the saved affine matrix.
        """

        io_it = self.get_io_iterator()

        for dwi, bval, bvec, omoved, oafffine in io_it:
            # Load the data from the input files and store into objects.
            logging.info(f"Loading {dwi}")
            data, affine = load_nifti(dwi)

            bvals, bvecs = read_bvals_bvecs(bval, bvec)
            # If all b-values are smaller or equal to the b0 threshold, it is
            # assumed that no thresholding is requested
            if any(mask_non_weighted_bvals(bvals, b0_threshold)):
                if b0_threshold < bvals.min():
                    warn(
                        f"b0_threshold (value: {b0_threshold}) is too low, "
                        "increase your b0_threshold. It should be higher than the "
                        f"lowest b0 value ({bvals.min()}).",
                        stacklevel=2,
                    )
            gtab = gradient_table(
                bvals, bvecs=bvecs, b0_threshold=b0_threshold, atol=bvecs_tol
            )

            reg_img, reg_affines = motion_correction(
                data=data, gtab=gtab, affine=affine
            )

            # Saving the corrected image file
            save_nifti(omoved, reg_img.get_fdata(), affine)
            # Write the affine matrix array to disk
            with open(oafffine, "w") as outfile:
                outfile.write(f"# Array shape: {reg_affines.shape}\n")
                for affine_slice in reg_affines:
                    np.savetxt(outfile, affine_slice, fmt="%-7.2f")
                    outfile.write("# New slice\n")


class BundleWarpFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "bundlewarp"

    def run(
        self,
        static_file,
        moving_file,
        dist=None,
        alpha=0.3,
        beta=20,
        max_iter=15,
        affine=True,
        out_dir="",
        out_linear_moved="linearly_moved.trk",
        out_nonlinear_moved="nonlinearly_moved.trk",
        out_warp_transform="warp_transform.npy",
        out_warp_kernel="warp_kernel.npy",
        out_dist="distance_matrix.npy",
        out_matched_pairs="matched_pairs.npy",
    ):
        """BundleWarp: streamline-based nonlinear registration.

        BundleWarp :footcite:p:`Chandio2023` is a nonrigid registration method
        for deformable registration of white matter tracts.

        Parameters
        ----------
        static_file : string
            Path to the static (reference) .trk file.
        moving_file : string
            Path to the moving (target to be registered) .trk file.
        dist : string, optional
            Path to the precalculated distance matrix file.
        alpha : float, optional
            Represents the trade-off between regularizing the deformation and
            having points match very closely. Lower value of alpha means high
            deformations. It is represented with λ in BundleWarp paper. NOTE:
            setting alpha<=0.01 will result in highly deformable registration
            that could extremely modify the original anatomy of the moving
            bundle.
        beta : int, optional
            Represents the strength of the interaction between points
            Gaussian kernel size.
        max_iter : int, optional
            Maximum number of iterations for deformation process in ml-CPD
            method.
        affine : boolean, optional
            If False, use rigid registration as starting point. (default True)
        out_dir : string, optional
            Output directory.
        out_linear_moved : string, optional
            Filename of linearly moved bundle.
        out_nonlinear_moved : string, optional
            Filename of nonlinearly moved (warped) bundle.
        out_warp_transform : string, optional
            Filename of warp transformations generated by BundleWarp.
        out_warp_kernel : string, optional
            Filename of regularization gaussian kernel generated by BundleWarp.
        out_dist : string, optional
            Filename of MDF distance matrix.
        out_matched_pairs : string, optional
            Filename of matched pairs; streamline correspondences between two
            bundles.

        References
        ----------
        .. footbibliography::
        """

        logging.info(f"Loading static file {static_file}")
        logging.info(f"Loading moving file {moving_file}")

        static_obj = nib.streamlines.load(static_file)
        moving_obj = nib.streamlines.load(moving_file)

        static, _ = static_obj.streamlines, static_obj.header
        moving, moving_header = moving_obj.streamlines, moving_obj.header

        static = set_number_of_points(static, nb_points=20)
        moving = set_number_of_points(moving, nb_points=20)

        deformed_bundle, affine_bundle, dists, mp, warp = bundlewarp(
            static,
            moving,
            dist=dist,
            alpha=alpha,
            beta=beta,
            max_iter=max_iter,
            affine=affine,
        )

        logging.info(f"Saving output file {out_linear_moved}")
        new_tractogram = nib.streamlines.Tractogram(
            affine_bundle, affine_to_rasmm=np.eye(4)
        )
        nib.streamlines.save(
            new_tractogram, pjoin(out_dir, out_linear_moved), header=moving_header
        )

        logging.info(f"Saving output file {out_nonlinear_moved}")
        new_tractogram = nib.streamlines.Tractogram(
            deformed_bundle, affine_to_rasmm=np.eye(4)
        )
        nib.streamlines.save(
            new_tractogram, pjoin(out_dir, out_nonlinear_moved), header=moving_header
        )

        logging.info(f"Saving output file {out_warp_transform}")
        np.save(pjoin(out_dir, out_warp_transform), np.array(warp["transforms"]))

        logging.info(f"Saving output file {out_warp_kernel}")
        np.save(pjoin(out_dir, out_warp_kernel), np.array(warp["gaussian_kernel"]))

        logging.info(f"Saving output file {out_dist}")
        np.save(pjoin(out_dir, out_dist), dist)

        logging.info(f"Saving output file {out_matched_pairs}")
        np.save(pjoin(out_dir, out_matched_pairs), mp)



"""
===================================
Brain segmentation with median_otsu
===================================

We show how to extract brain information and mask from a b0 image using DIPY_'s
``segment.mask`` module.

First import the necessary modules:
"""

import matplotlib.pyplot as plt
import numpy as np

from dipy.core.histeq import histeq
from dipy.data import get_fnames
from dipy.io.image import load_nifti, save_nifti
from dipy.segment.mask import median_otsu

###############################################################################
# Download and read the data for this tutorial.
#
# The ``scil_b0`` dataset contains different data from different companies and
# models. For this example, the data comes from a 1.5 Tesla Siemens MRI.

data_fnames = get_fnames(name="scil_b0")
data, affine = load_nifti(data_fnames[1])
data = np.squeeze(data)

###############################################################################
# Segment the brain using DIPY's ``mask`` module.
#
# ``median_otsu`` returns the segmented brain data and a binary mask of the
# brain. It is possible to fine tune the parameters of ``median_otsu``
# (``median_radius`` and ``num_pass``) if extraction yields incorrect results
# but the default parameters work well on most volumes. For this example,
# we used 2 as ``median_radius`` and 1 as ``num_pass``

b0_mask, mask = median_otsu(data, median_radius=2, numpass=1)

###############################################################################
# Saving the segmentation results is very easy. We need the ``b0_mask``, and
# the binary mask volumes. The affine matrix which transform the image's
# coordinates to the world coordinates is also needed. Here, we choose to save
# both images in ``float32``.

fname = "se_1.5t"
save_nifti(fname + "_binary_mask.nii.gz", mask.astype(np.float32), affine)
save_nifti(fname + "_mask.nii.gz", b0_mask.astype(np.float32), affine)

###############################################################################
# Quick view of the results middle slice using ``matplotlib``.

sli = data.shape[2] // 2
plt.figure("Brain segmentation")
plt.subplot(1, 2, 1).set_axis_off()
plt.imshow(histeq(data[:, :, sli].astype("float")).T, cmap="gray", origin="lower")

plt.subplot(1, 2, 2).set_axis_off()
plt.imshow(histeq(b0_mask[:, :, sli].astype("float")).T, cmap="gray", origin="lower")
plt.savefig(f"{fname}_median_otsu.png", bbox_inches="tight")

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# An application of median_otsu for brain segmentation.
#
#
# ``median_otsu`` can also automatically crop the outputs to remove the largest
# possible number of background voxels. This makes outputted data significantly
# smaller. Auto-cropping in ``median_otsu`` is activated by setting the
# ``autocrop`` parameter to ``True``.

b0_mask_crop, mask_crop = median_otsu(data, median_radius=4, numpass=4, autocrop=True)

###############################################################################
# Saving cropped data as demonstrated previously.

save_nifti(fname + "_binary_mask_crop.nii.gz", mask_crop.astype(np.float32), affine)
save_nifti(fname + "_mask_crop.nii.gz", b0_mask_crop.astype(np.float32), affine)



"""Utility functions for file formats"""

import logging
import numbers
import os

import nibabel as nib
from nibabel import Nifti1Image
from nibabel.streamlines import detect_format
import numpy as np
from trx import trx_file_memmap

import dipy
from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

pd, have_pd, _ = optional_package("pandas")

if have_pd:
    import pandas as pd


def nifti1_symmat(image_data, *args, **kwargs):
    """Returns a Nifti1Image with a symmetric matrix intent

    Parameters
    ----------
    image_data : array-like
        should have lower triangular elements of a symmetric matrix along the
        last dimension
    *args
        Passed to Nifti1Image
    **kwargs
        Passed to Nifti1Image

    Returns
    -------
    image : Nifti1Image
        5d, extra dimensions added before the last. Has symmetric matrix intent
        code

    """
    image_data = make5d(image_data)
    last_dim = image_data.shape[-1]
    n = (np.sqrt(1 + 8 * last_dim) - 1) / 2
    if (n % 1) != 0:
        raise ValueError("input_data does not seem to have matrix elements")

    image = Nifti1Image(image_data, *args, **kwargs)
    hdr = image.header
    hdr.set_intent("symmetric matrix", (n,))
    return image


def make5d(data):
    """reshapes the input to have 5 dimensions, adds extra dimensions just
    before the last dimension
    """
    data = np.asarray(data)
    if data.ndim > 5:
        raise ValueError("input is already more than 5d")
    shape = data.shape
    shape = shape[:-1] + (1,) * (5 - len(shape)) + shape[-1:]
    return data.reshape(shape)


@warning_for_keywords()
def decfa(img_orig, *, scale=False):
    """
    Create a nifti-compliant directional-encoded color FA image.

    Parameters
    ----------
    img_orig : Nifti1Image class instance.
        Contains encoding of the DEC FA image with a 4D volume of data, where
        the elements on the last dimension represent R, G and B components.

    scale: bool.
        Whether to scale the incoming data from the 0-1 to the 0-255 range
        expected in the output.

    Returns
    -------
    img : Nifti1Image class instance with dtype set to store tuples of
        uint8 in (R, G, B) order.


    Notes
    -----
    For a description of this format, see:

    https://nifti.nimh.nih.gov/nifti-1/documentation/nifti1fields/nifti1fields_pages/datatype.html
    """

    dest_dtype = np.dtype([("R", "uint8"), ("G", "uint8"), ("B", "uint8")])
    out_data = np.zeros(img_orig.shape[:3], dtype=dest_dtype)

    data_orig = np.asanyarray(img_orig.dataobj)

    if scale:
        data_orig = (data_orig * 255).astype("uint8")

    for ii in np.ndindex(img_orig.shape[:3]):
        val = data_orig[ii]
        out_data[ii] = (val[0], val[1], val[2])

    new_hdr = img_orig.header
    new_hdr["dim"][4] = 1
    new_hdr.set_intent(1001, name="Color FA")
    new_hdr.set_data_dtype(dest_dtype)

    return Nifti1Image(out_data, affine=img_orig.affine, header=new_hdr)


def decfa_to_float(img_orig):
    """
    Convert a nifti-compliant directional-encoded color FA image into a
    nifti image with RGB encoded in floating point resolution.

    Parameters
    ----------
    img_orig : Nifti1Image class instance.
        Contains encoding of the DEC FA image with a 3D volume of data, where
        each element is a (R, G, B) tuple in uint8.

    Returns
    -------
    img : Nifti1Image class instance with float dtype.

    Notes
    -----
    For a description of this format, see:

    https://nifti.nimh.nih.gov/nifti-1/documentation/nifti1fields/nifti1fields_pages/datatype.html
    """

    data_orig = np.asanyarray(img_orig.dataobj)
    out_data = np.zeros(data_orig.shape + (3,), dtype=np.uint8)

    for ii in np.ndindex(img_orig.shape[:3]):
        val = data_orig[ii]
        out_data[ii] = np.array([val[0], val[1], val[2]])

    new_hdr = img_orig.header
    new_hdr["dim"][4] = 3

    # Remove the original intent
    new_hdr.set_intent(0)
    new_hdr.set_data_dtype(float)

    return Nifti1Image(out_data, affine=img_orig.affine, header=new_hdr)


def is_reference_info_valid(affine, dimensions, voxel_sizes, voxel_order):
    """Validate basic data type and value of spatial attribute.

    Does not ensure that voxel_sizes and voxel_order are self-coherent with
    the affine.

    Only verifies the following:
        - affine is of the right type (float) and dimension (4,4)
        - affine contain values in the rotation part
        - dimensions is of right type (int) and length (3)
        - voxel_sizes is of right type (float) and length (3)
        - voxel_order is of right type (str) and length (3)

    The listed parameters are what is expected, provide something else and this
    function should fail (cover common mistakes).

    Parameters
    ----------
    affine: ndarray (4,4)
        Transformation of VOX to RASMM
    dimensions: ndarray (3,), int16
        Volume shape for each axis
    voxel_sizes:  ndarray (3,), float32
        Size of voxel for each axis
    voxel_order: string
        Typically 'RAS' or 'LPS'

    Returns
    -------
    output : bool
        Does the input represent a valid 'state' of spatial attribute

    """
    all_valid = True
    only_3d_warning = False

    if not affine.shape == (4, 4):
        all_valid = False
        logging.warning("Transformation matrix must be 4x4")

    if not affine[0:3, 0:3].any():
        all_valid = False
        logging.warning("Rotation matrix cannot be all zeros")

    if not len(dimensions) >= 3:
        all_valid = False
        only_3d_warning = True

    for i in dimensions:
        if not isinstance(i, numbers.Integral):
            all_valid = False
            logging.warning("Dimensions must be int.")
        if i <= 0:
            all_valid = False
            logging.warning("Dimensions must be above 0.")

    if not len(voxel_sizes) >= 3:
        all_valid = False
        only_3d_warning = True
    for i in voxel_sizes:
        if not isinstance(i, numbers.Number):
            all_valid = False
            logging.warning("Voxel size must be int/float.")
        if i <= 0:
            all_valid = False
            logging.warning("Voxel size must be above 0.")

    if not len(voxel_order) >= 3:
        all_valid = False
        only_3d_warning = True
    for i in voxel_order:
        if not isinstance(i, str):
            all_valid = False
            logging.warning("Voxel order must be string/char.")
        if i not in ["R", "A", "S", "L", "P", "I"]:
            all_valid = False
            logging.warning("Voxel order does not follow convention.")

    if only_3d_warning:
        logging.warning("Only 3D (and above) reference are considered valid.")

    return all_valid


def split_name_with_gz(filename):
    """
    Returns the clean basename and extension of a file.
    Means that this correctly manages the ".nii.gz" extensions.

    Parameters
    ----------
    filename: str
        The filename to clean

    Returns
    -------
        base, ext : tuple(str, str)
        Clean basename and the full extension
    """
    base, ext = os.path.splitext(filename)

    if ext.lower() == ".gz":
        # Test if we have a .nii additional extension
        temp_base, add_ext = os.path.splitext(base)

        if add_ext.lower() == ".nii" or add_ext.lower() == ".trk":
            ext = add_ext + ext
            base = temp_base

    return base, ext


def get_reference_info(reference):
    """Will compare the spatial attribute of 2 references.

    Parameters
    ----------
    reference : Nifti or Trk filename, Nifti1Image or TrkFile, Nifti1Header or
        trk.header (dict), TrxFile or trx.header (dict)
        Reference that provides the spatial attribute.

    Returns
    -------
    output : tuple
        - affine ndarray (4,4), np.float32, transformation of VOX to RASMM
        - dimensions ndarray (3,), int16, volume shape for each axis
        - voxel_sizes  ndarray (3,), float32, size of voxel for each axis
        - voxel_order, string, Typically 'RAS' or 'LPS'
    """

    is_nifti = False
    is_trk = False
    is_sft = False
    is_trx = False
    if isinstance(reference, str):
        _, ext = split_name_with_gz(reference)
        ext = ext.lower()
        if ext in [".nii", ".nii.gz"]:
            header = nib.load(reference).header
            is_nifti = True
        elif ext == ".trk":
            header = nib.streamlines.load(reference, lazy_load=True).header
            is_trk = True
        elif ext == ".trx":
            header = trx_file_memmap.load(reference).header
            is_trx = True
    elif isinstance(reference, trx_file_memmap.TrxFile):
        header = reference.header
        is_trx = True
    elif isinstance(reference, nib.nifti1.Nifti1Image):
        header = reference.header
        is_nifti = True
    elif isinstance(reference, nib.streamlines.trk.TrkFile):
        header = reference.header
        is_trk = True
    elif isinstance(reference, nib.nifti1.Nifti1Header):
        header = reference
        is_nifti = True
    elif isinstance(reference, dict) and "magic_number" in reference:
        header = reference
        is_trk = True
    elif isinstance(reference, dict) and "NB_VERTICES" in reference:
        header = reference
        is_trx = True
    elif isinstance(reference, dipy.io.stateful_tractogram.StatefulTractogram):
        is_sft = True

    if is_nifti:
        affine = header.get_best_affine()
        dimensions = header["dim"][1:4]
        voxel_sizes = header["pixdim"][1:4]

        if not affine[0:3, 0:3].any():
            raise ValueError(
                "Invalid affine, contains only zeros."
                "Cannot determine voxel order from transformation"
            )
        voxel_order = "".join(nib.aff2axcodes(affine))
    elif is_trk:
        affine = header["voxel_to_rasmm"]
        dimensions = header["dimensions"]
        voxel_sizes = header["voxel_sizes"]
        voxel_order = header["voxel_order"]
    elif is_sft:
        affine, dimensions, voxel_sizes, voxel_order = reference.space_attributes
    elif is_trx:
        affine = header["VOXEL_TO_RASMM"]
        dimensions = header["DIMENSIONS"]
        voxel_sizes = nib.affines.voxel_sizes(affine)
        voxel_order = "".join(nib.aff2axcodes(affine))
    else:
        raise TypeError("Input reference is not one of the supported format")

    if isinstance(voxel_order, np.bytes_):
        voxel_order = voxel_order.decode("utf-8")

    is_reference_info_valid(affine, dimensions, voxel_sizes, voxel_order)

    return affine.astype(np.float32), dimensions, voxel_sizes, voxel_order


def is_header_compatible(reference_1, reference_2):
    """Will compare the spatial attribute of 2 references

    Parameters
    ----------
    reference_1 : Nifti or Trk filename, Nifti1Image or TrkFile,
        Nifti1Header or trk.header (dict)
        Reference that provides the spatial attribute.
    reference_2 : Nifti or Trk filename, Nifti1Image or TrkFile,
        Nifti1Header or trk.header (dict)
        Reference that provides the spatial attribute.

    Returns
    -------
    output : bool
        Does all the spatial attribute match
    """

    affine_1, dimensions_1, voxel_sizes_1, voxel_order_1 = get_reference_info(
        reference_1
    )
    affine_2, dimensions_2, voxel_sizes_2, voxel_order_2 = get_reference_info(
        reference_2
    )

    identical_header = True
    if not np.allclose(affine_1, affine_2, rtol=1e-03, atol=1e-03):
        logging.error("Affine not equal")
        identical_header = False

    if not np.array_equal(dimensions_1, dimensions_2):
        logging.error("Dimensions not equal")
        identical_header = False

    if not np.allclose(voxel_sizes_1, voxel_sizes_2, rtol=1e-03, atol=1e-03):
        logging.error("Voxel_size not equal")
        identical_header = False

    if voxel_order_1 != voxel_order_2:
        logging.error("Voxel_order not equal")
        identical_header = False

    return identical_header


def create_tractogram_header(
    tractogram_type, affine, dimensions, voxel_sizes, voxel_order
):
    """Write a standard trk/tck header from spatial attribute"""
    if isinstance(tractogram_type, str):
        tractogram_type = detect_format(tractogram_type)

    new_header = tractogram_type.create_empty_header()
    new_header[nib.streamlines.Field.VOXEL_SIZES] = tuple(voxel_sizes)
    new_header[nib.streamlines.Field.DIMENSIONS] = tuple(dimensions)
    new_header[nib.streamlines.Field.VOXEL_TO_RASMM] = affine
    new_header[nib.streamlines.Field.VOXEL_ORDER] = voxel_order

    return new_header


def create_nifti_header(affine, dimensions, voxel_sizes):
    """Write a standard nifti header from spatial attribute"""
    new_header = nib.Nifti1Header()
    new_header.set_sform(affine)
    new_header["dim"][1:4] = dimensions
    new_header["pixdim"][1:4] = voxel_sizes

    new_header.affine = new_header.get_best_affine()

    return new_header


@warning_for_keywords()
def save_buan_profiles_hdf5(fname, dt, *, key=None):
    """Saves the given input dataframe to .h5 file

    Parameters
    ----------
    fname : string
        file name for saving the hdf5 file
    dt : Pandas DataFrame
        DataFrame to be saved as .h5 file
    key : str, optional
        Key to retrieve the contents in the HDF5 file. The file rootname will
        be used if not provided.

    """

    df = pd.DataFrame(dt)
    filename_hdf5 = fname + ".h5"

    if key is None:
        base_name_parts, _ = os.path.splitext(os.path.basename(fname))
        key = base_name_parts.split(".")[0]

    store = pd.HDFStore(filename_hdf5, complevel=9)
    store.append(key, df, data_columns=True, complevel=9)
    store.close()


@warning_for_keywords()
def read_img_arr_or_path(data, *, affine=None):
    """
    Helper function that handles inputs that can be paths, nifti img or arrays

    Parameters
    ----------
    data : array or nib.Nifti1Image or str.
        Either as a 3D/4D array or as a nifti image object, or as
        a string containing the full path to a nifti file.

    affine : 4x4 array, optional.
        Must be provided for `data` provided as an array. If provided together
        with Nifti1Image or str `data`, this input will over-ride the affine
        that is stored in the `data` input. Default: use the affine stored
        in `data`.

    Returns
    -------
    data, affine : ndarray and 4x4 array
    """
    if isinstance(data, np.ndarray) and affine is None:
        raise ValueError(
            "If data is provided as an array, an affine has ", "to be provided as well"
        )
    if isinstance(data, str):
        data = nib.load(data)
    if isinstance(data, nib.Nifti1Image):
        if affine is None:
            affine = data.affine
        data = data.get_fdata()
    return data, affine



cimport numpy as cnp

from dipy.tracking.stopping_criterion cimport StoppingCriterion, StreamlineStatus
from dipy.direction.pmf cimport PmfGen
from dipy.tracking.tracker_parameters cimport TrackerParameters


cdef void generate_tractogram_c(double[:,::1] seed_positions,
                                double[:,::1] seed_directions,
                                int nbr_threads,
                                StoppingCriterion sc,
                                TrackerParameters params,
                                PmfGen pmf_gen,
                                double** streamlines,
                                int* length,
                                StreamlineStatus* status)


cdef StreamlineStatus generate_local_streamline(double* seed,
                                   double* position,
                                   double* stream,
                                   int* stream_idx,
                                   StoppingCriterion sc,
                                   TrackerParameters params,
                                   PmfGen pmf_gen) noexcept nogil


cdef void prepare_pmf(double* pmf,
                      double* point,
                      PmfGen pmf_gen,
                      double pmf_threshold,
                      int pmf_len) noexcept nogil



"""
======================================================
Visualization of ROI Surface Rendered with Streamlines
======================================================

Here is a simple tutorial following the probabilistic CSA Tracking Example in
which we generate a dataset of streamlines from a corpus callosum ROI, and
then display them with the seed ROI rendered in 3D with 50% transparency.

Let's start by importing the relevant modules.
"""

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames
from dipy.direction import peaks_from_model
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data
from dipy.reconst.shm import CsaOdfModel
from dipy.tracking import utils
from dipy.tracking.local_tracking import LocalTracking
from dipy.tracking.stopping_criterion import ThresholdStoppingCriterion
from dipy.tracking.streamline import Streamlines
from dipy.viz import actor, colormap as cmap, window

###############################################################################
# First, we need to generate some streamlines. For a more complete
# description of these steps, please refer to the CSA Probabilistic Tracking
# Tutorial.

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")
label_fname = get_fnames(name="stanford_labels")

data, affine, hardi_img = load_nifti(hardi_fname, return_img=True)
labels = load_nifti_data(label_fname)
bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)

white_matter = (labels == 1) | (labels == 2)

csa_model = CsaOdfModel(gtab, sh_order_max=6)
csa_peaks = peaks_from_model(
    csa_model,
    data,
    default_sphere,
    relative_peak_threshold=0.8,
    min_separation_angle=45,
    mask=white_matter,
)

stopping_criterion = ThresholdStoppingCriterion(csa_peaks.gfa, 0.25)

seed_mask = labels == 2
seeds = utils.seeds_from_mask(seed_mask, affine, density=[1, 1, 1])

# Initialization of LocalTracking. The computation happens in the next step.
streamlines = LocalTracking(csa_peaks, stopping_criterion, seeds, affine, step_size=2)

# Compute streamlines and store as a list.
streamlines = Streamlines(streamlines)

###############################################################################
# We will create a streamline actor from the streamlines.

streamlines_actor = actor.line(streamlines, colors=cmap.line_colors(streamlines))

###############################################################################
# Next, we create a surface actor from the corpus callosum seed ROI. We
# provide the ROI data, the affine, the color in [R,G,B], and the opacity as
# a decimal between zero and one. Here, we set the color as blue/green with
# 50% opacity.

surface_opacity = 0.5
surface_color = [0, 1, 1]

seedroi_actor = actor.contour_from_roi(
    seed_mask, affine=affine, color=surface_color, opacity=surface_opacity
)

###############################################################################
# Next, we initialize a ''Scene'' object and add both actors
# to the rendering.

scene = window.Scene()
scene.add(streamlines_actor)
scene.add(seedroi_actor)

###############################################################################
# If you uncomment the following line, the rendering will pop up in an
# interactive window.

interactive = False
if interactive:
    window.show(scene)

window.record(scene=scene, out_path="contour_from_roi_tutorial.png", size=(1200, 900))

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# A top view of corpus callosum streamlines with the blue transparent
# seed ROI in the center.



"""
===========================================
Tractography Clustering - Available Metrics
===========================================

This page lists available metrics that can be used by the tractography
clustering framework. For every metric a brief description is provided
explaining: what it does, when it's useful and how to use it. If you are not
familiar with the tractography clustering framework, check this tutorial
:ref:`clustering-framework`. See :footcite:p:`Garyfallidis2012a` for more
information on the metrics.

.. contents:: Available Metrics
    :local:
    :depth: 1

Let's start by importing the necessary modules.
"""

import numpy as np

from dipy.segment.clustering import QuickBundles
from dipy.segment.featurespeed import VectorOfEndpointsFeature
from dipy.segment.metric import (
    AveragePointwiseEuclideanMetric,
    CosineMetric,
    SumPointwiseEuclideanMetric,
)
from dipy.tracking.streamline import set_number_of_points
from dipy.viz import actor, colormap, window

###############################################################################
# .. note::
#
#     All examples assume a function `get_streamlines` exists. We defined here
#     a simple function to do so. It imports the necessary modules and loads a
#     small streamline bundle.


def get_streamlines():
    from dipy.data import get_fnames
    from dipy.io.streamline import load_tractogram

    fname = get_fnames(name="fornix")
    fornix = load_tractogram(fname, "same", bbox_valid_check=False)

    return fornix.streamlines


###############################################################################
# .. _clustering-examples-AveragePointwiseEuclideanMetric:
#
# Average of Pointwise Euclidean Metric
# =====================================
# **What:** Instances of `AveragePointwiseEuclideanMetric` first compute the
# pointwise Euclidean distance between two sequences *of same length* then
# return the average of those distances. This metric takes as inputs two
# features that are sequences containing the same number of elements.
#
# **When:** By default the `QuickBundles` clustering will resample your
# streamlines on-the-fly so they have 12 points. If for some reason you want
# to avoid this and you made sure all your streamlines already have the same
# number of points, you can manually provide an instance of
# `AveragePointwiseEuclideanMetric` to `QuickBundles`. Since the default
# `Feature` is the `IdentityFeature` the streamlines won't be resampled thus
# saving some computational time.
#
# **Note:** Inputs must be sequences of the same length.


# Get some streamlines.
streamlines = get_streamlines()  # Previously defined.

# Make sure our streamlines have the same number of points.
streamlines = set_number_of_points(streamlines, nb_points=12)

# Create the instance of `AveragePointwiseEuclideanMetric` to use.
metric = AveragePointwiseEuclideanMetric()
qb = QuickBundles(threshold=10.0, metric=metric)
clusters = qb.cluster(streamlines)

print("Nb. clusters:", len(clusters))
print("Cluster sizes:", map(len, clusters))

###############################################################################
# .. _clustering-examples-SumPointwiseEuclideanMetric:
#
# Sum of Pointwise Euclidean Metric
# =================================
# **What:** Instances of `SumPointwiseEuclideanMetric` first compute the
# pointwise Euclidean distance between two sequences *of same length* then
# return the sum of those distances.
#
# **When:** This metric mainly exists because it is used internally by
# `AveragePointwiseEuclideanMetric`.
#
# **Note:** Inputs must be sequences of the same length.

# Get some streamlines.
streamlines = get_streamlines()  # Previously defined.

# Make sure our streamlines have the same number of points.
nb_points = 12
streamlines = set_number_of_points(streamlines, nb_points=nb_points)

# Create the instance of `SumPointwiseEuclideanMetric` to use.
metric = SumPointwiseEuclideanMetric()
qb = QuickBundles(threshold=10.0 * nb_points, metric=metric)
clusters = qb.cluster(streamlines)

print("Nb. clusters:", len(clusters))
print("Cluster sizes:", map(len, clusters))

###############################################################################
# Cosine Metric
# =============
# **What:** Instances of `CosineMetric` compute the cosine distance between two
# vectors (for more information see the
# `wiki page <https://en.wikipedia.org/wiki/Cosine_similarity>`_).
#
# **When:** This metric can be useful when you *only* need information about
# the orientation of a streamline.
#
# **Note:** Inputs must be vectors (i.e. 1D array).

# Enables/disables interactive visualization
interactive = False

# Get some streamlines.
streamlines = get_streamlines()  # Previously defined.

feature = VectorOfEndpointsFeature()
metric = CosineMetric(feature)
qb = QuickBundles(threshold=0.1, metric=metric)
clusters = qb.cluster(streamlines)

# Color each streamline according to the cluster they belong to.
colormap = colormap.create_colormap(np.arange(len(clusters)))
colormap_full = np.ones((len(streamlines), 3))
for cluster, color in zip(clusters, colormap):
    colormap_full[cluster.indices] = color

# Visualization
scene = window.Scene()
scene.clear()
scene.SetBackground(0, 0, 0)
scene.add(actor.streamtube(streamlines, colors=colormap_full))
window.record(scene=scene, out_path="cosine_metric.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the streamlines colored according to their orientation.
#
#
# References
# ----------
#
# .. footbibliography::
#



"""

=================================================
Between-volumes Motion Correction on DWI datasets
=================================================

During a dMRI acquisition, the subject motion inevitable. This motion implies
a misalignment between N volumes on a dMRI dataset. A common way to solve this
issue is to perform a registration on each acquired volume to a
reference b = 0 :footcite:p:`Jenkinson2001`.

This preprocessing is a highly recommended step that should be executed before
any dMRI dataset analysis.


Let's import some essential functions.
"""

from dipy.align import motion_correction
from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti

###############################################################################
# We choose one of the data from the datasets in dipy_. However, you can
# replace the following line with the path of your image.

dwi_fname, dwi_bval_fname, dwi_bvec_fname = get_fnames(name="sherbrooke_3shell")

###############################################################################
# We load the image and the affine of the image. The affine is the
# transformation matrix which maps image coordinates to world (mm) coordinates.
# We also load the b-values and b-vectors.

data, affine = load_nifti(dwi_fname)
bvals, bvecs = read_bvals_bvecs(dwi_bval_fname, dwi_bvec_fname)

###############################################################################
# This data has 193 volumes. For this demo purpose, we decide to reduce the
# number of volumes to 3. However, we do not recommend to perform a motion
# correction with less than 10 volumes.

data_small = data[..., :3]
bvals_small = bvals[:3]
bvecs_small = bvecs[:3]
gtab = gradient_table(bvals_small, bvecs=bvecs_small)

###############################################################################
# Start motion correction of our reduced DWI dataset(between-volumes motion
# correction).

data_corrected, reg_affines = motion_correction(data_small, gtab, affine=affine)

###############################################################################
# Save our DWI dataset corrected to a new Nifti file.

save_nifti(
    "motion_correction.nii.gz", data_corrected.get_fdata(), data_corrected.affine
)

###############################################################################
# References
# ----------
#
# .. footbibliography::
#



# Tests for FORECAST fitting and metrics

import warnings

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal
import pytest

from dipy.core.sphere_stats import angular_similarity
from dipy.data import default_sphere, get_3shell_gtab, get_sphere
from dipy.direction.peaks import peak_directions
from dipy.reconst.forecast import ForecastModel
from dipy.reconst.shm import descoteaux07_legacy_msg
from dipy.sims.voxel import multi_tensor
from dipy.utils.optpkg import optional_package

cvxpy, have_cvxpy, _ = optional_package("cvxpy", min_version="1.4.1")
needs_cvxpy = pytest.mark.skipif(not have_cvxpy, reason="Requires CVXPY")


# Object to hold module global data
class _C:
    pass


data = _C()


def setup_module():
    global data
    data.gtab = get_3shell_gtab()
    data.mevals = np.array(([0.0017, 0.0003, 0.0003], [0.0017, 0.0003, 0.0003]))
    data.angl = [(0, 0), (60, 0)]
    data.S, data.sticks = multi_tensor(
        data.gtab, data.mevals, S0=100.0, angles=data.angl, fractions=[50, 50], snr=None
    )
    data.sh_order_max = 6
    data.lambda_lb = 1e-8
    data.lambda_csd = 1.0
    sphere = get_sphere(name="repulsion100")
    data.sphere = sphere.vertices[0 : int(sphere.vertices.shape[0] / 2), :]


@needs_cvxpy
def test_forecast_positive_constrain():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(
            data.gtab,
            sh_order_max=data.sh_order_max,
            lambda_lb=data.lambda_lb,
            dec_alg="POS",
            sphere=data.sphere,
        )
    f_fit = fm.fit(data.S)

    sphere = get_sphere(name="repulsion100")
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fodf = f_fit.odf(sphere, clip_negative=False)
    assert_almost_equal(fodf[fodf < 0].sum(), 0, 2)

    coeff = f_fit.sh_coeff
    c0 = np.sqrt(1.0 / (4 * np.pi))
    assert_almost_equal(coeff[0], c0, 5)


def test_forecast_csd():
    sphere = get_sphere(name="repulsion100")
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(
            data.gtab, dec_alg="CSD", sphere=data.sphere, lambda_csd=data.lambda_csd
        )
    f_fit = fm.fit(data.S)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fodf_csd = f_fit.odf(sphere, clip_negative=False)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(
            data.gtab,
            sh_order_max=data.sh_order_max,
            lambda_lb=data.lambda_lb,
            dec_alg="WLS",
        )
    f_fit = fm.fit(data.S)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fodf_wls = f_fit.odf(sphere, clip_negative=False)

    value = fodf_wls[fodf_wls < 0].sum() < fodf_csd[fodf_csd < 0].sum()
    assert_equal(value, 1)


def test_forecast_odf():
    # check FORECAST fODF at different SH order
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(data.gtab, sh_order_max=4, dec_alg="CSD", sphere=data.sphere)
    f_fit = fm.fit(data.S)
    sphere = default_sphere
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fodf = f_fit.odf(sphere)
    directions, _, _ = peak_directions(
        fodf, sphere, relative_peak_threshold=0.35, min_separation_angle=25
    )
    assert_equal(len(directions), 2)
    assert_almost_equal(angular_similarity(directions, data.sticks), 2, 1)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(data.gtab, sh_order_max=6, dec_alg="CSD", sphere=data.sphere)
    f_fit = fm.fit(data.S)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fodf = f_fit.odf(sphere)
    directions, _, _ = peak_directions(
        fodf, sphere, relative_peak_threshold=0.35, min_separation_angle=25
    )
    assert_equal(len(directions), 2)
    assert_almost_equal(angular_similarity(directions, data.sticks), 2, 1)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(data.gtab, sh_order_max=8, dec_alg="CSD", sphere=data.sphere)
    f_fit = fm.fit(data.S)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fodf = f_fit.odf(sphere)
    directions, _, _ = peak_directions(
        fodf, sphere, relative_peak_threshold=0.35, min_separation_angle=25
    )
    assert_equal(len(directions), 2)
    assert_almost_equal(angular_similarity(directions, data.sticks), 2, 1)

    # stronger regularization is required for high order SH
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(
            data.gtab, sh_order_max=10, dec_alg="CSD", sphere=sphere.vertices
        )
    f_fit = fm.fit(data.S)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fodf = f_fit.odf(sphere)
    directions, _, _ = peak_directions(
        fodf, sphere, relative_peak_threshold=0.35, min_separation_angle=25
    )
    assert_equal(len(directions), 2)
    assert_almost_equal(angular_similarity(directions, data.sticks), 2, 1)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(
            data.gtab, sh_order_max=12, dec_alg="CSD", sphere=sphere.vertices
        )
    f_fit = fm.fit(data.S)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fodf = f_fit.odf(sphere)
    directions, _, _ = peak_directions(
        fodf, sphere, relative_peak_threshold=0.35, min_separation_angle=25
    )
    assert_equal(len(directions), 2)
    assert_almost_equal(angular_similarity(directions, data.sticks), 2, 1)


def test_forecast_indices():
    # check anisotropic tensor
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(
            data.gtab, sh_order_max=2, lambda_lb=data.lambda_lb, dec_alg="WLS"
        )
    f_fit = fm.fit(data.S)

    d_par = f_fit.dpar
    d_perp = f_fit.dperp

    assert_almost_equal(d_par, data.mevals[0, 0], 5)
    assert_almost_equal(d_perp, data.mevals[0, 1], 5)

    gt_fa = np.sqrt(
        0.5
        * (2 * (data.mevals[0, 0] - data.mevals[0, 1]) ** 2)
        / (data.mevals[0, 0] ** 2 + 2 * data.mevals[0, 1] ** 2)
    )
    gt_md = (data.mevals[0, 0] + 2 * data.mevals[0, 1]) / 3.0

    assert_almost_equal(f_fit.fractional_anisotropy(), gt_fa, 2)
    assert_almost_equal(f_fit.mean_diffusivity(), gt_md, 5)

    # check isotropic tensor
    mevals = np.array(([0.003, 0.003, 0.003], [0.003, 0.003, 0.003]))
    data.angl = [(0, 0), (60, 0)]
    S, sticks = multi_tensor(
        data.gtab, mevals, S0=100.0, angles=data.angl, fractions=[50, 50], snr=None
    )

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(
            data.gtab,
            sh_order_max=data.sh_order_max,
            lambda_lb=data.lambda_lb,
            dec_alg="WLS",
        )
    f_fit = fm.fit(S)

    d_par = f_fit.dpar
    d_perp = f_fit.dperp

    assert_almost_equal(d_par, 3e-03, 5)
    assert_almost_equal(d_perp, 3e-03, 5)
    assert_almost_equal(f_fit.fractional_anisotropy(), 0.0, 5)
    assert_almost_equal(f_fit.mean_diffusivity(), 3e-03, 10)


def test_forecast_predict():
    # check anisotropic tensor
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(data.gtab, sh_order_max=8, dec_alg="CSD", sphere=data.sphere)
    f_fit = fm.fit(data.S)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        S = f_fit.predict(S0=1.0)

    mse = np.sum((S - data.S / 100.0) ** 2) / len(S)

    assert_almost_equal(mse, 0.0, 3)


def test_multivox_forecast():
    gtab = get_3shell_gtab()
    mevals = np.array(([0.0017, 0.0003, 0.0003], [0.0017, 0.0003, 0.0003]))

    angl1 = [(0, 0), (60, 0)]
    angl2 = [(90, 0), (45, 90)]
    angl3 = [(0, 0), (90, 0)]

    S = np.zeros((3, 1, 1, len(gtab.bvals)))
    S[0, 0, 0], _ = multi_tensor(
        gtab, mevals, S0=1.0, angles=angl1, fractions=[50, 50], snr=None
    )
    S[1, 0, 0], _ = multi_tensor(
        gtab, mevals, S0=1.0, angles=angl2, fractions=[50, 50], snr=None
    )
    S[2, 0, 0], _ = multi_tensor(
        gtab, mevals, S0=1.0, angles=angl3, fractions=[50, 50], snr=None
    )

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        fm = ForecastModel(gtab, sh_order_max=8, dec_alg="CSD")
    f_fit = fm.fit(S)

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        S_predict = f_fit.predict()

    assert_equal(S_predict.shape, S.shape)

    mse1 = np.sum((S_predict[0, 0, 0] - S[0, 0, 0]) ** 2) / len(gtab.bvals)
    assert_almost_equal(mse1, 0.0, 3)

    mse2 = np.sum((S_predict[1, 0, 0] - S[1, 0, 0]) ** 2) / len(gtab.bvals)
    assert_almost_equal(mse2, 0.0, 3)

    mse3 = np.sum((S_predict[2, 0, 0] - S[2, 0, 0]) ** 2) / len(gtab.bvals)
    assert_almost_equal(mse3, 0.0, 3)



"""
=======================================================
Tissue Classification of a T1-weighted Structural Image
=======================================================

This example explains how to segment a T1-weighted structural image by using
Bayesian formulation. The observation model (likelihood term) is defined as a
Gaussian distribution and a Markov Random Field (MRF) is used to model the
a priori probability of context-dependent patterns of different tissue
types of the brain. Expectation Maximization and Iterated Conditional
Modes are used to find the optimal solution. Similar algorithms have been
proposed by :footcite:t:`Zhang2001` and :footcite:p:`Avants2011` available
in FAST-FSL and ANTS-atropos, respectively.

Here we will use a T1-weighted image, that has been previously skull-stripped
and bias field corrected.
"""

import time

import matplotlib.pyplot as plt
import numpy as np

from dipy.data import get_fnames
from dipy.io.image import load_nifti_data
from dipy.segment.tissue import TissueClassifierHMRF

###############################################################################
# First we fetch the T1 volume from the Syn dataset and determine its shape.

t1_fname, _, _ = get_fnames(name="tissue_data")
t1 = load_nifti_data(t1_fname)
print(f"t1.shape {t1.shape}")

###############################################################################
# We have fetched the T1 volume. Now we will look at the axial and coronal
# slices of the image.

fig = plt.figure()
a = fig.add_subplot(1, 2, 1)
img_ax = np.rot90(t1[..., 89])
imgplot = plt.imshow(img_ax, cmap="gray")
a.axis("off")
a.set_title("Axial")
a = fig.add_subplot(1, 2, 2)
img_cor = np.rot90(t1[:, 128, :])
imgplot = plt.imshow(img_cor, cmap="gray")
a.axis("off")
a.set_title("Coronal")
plt.savefig("t1_image.png", bbox_inches="tight", pad_inches=0)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# T1-weighted image of healthy adult.
#
#
# Now we will define the other two parameters for the segmentation algorithm.
# We will segment three classes, namely corticospinal fluid (CSF), white matter
# (WM) and gray matter (GM).

nclass = 3

###############################################################################
# Then, the smoothness factor of the segmentation. Good performance is achieved
# with values between 0 and 0.5.

beta = 0.1

###############################################################################
# We could also set the number of iterations. By default this parameter is set
# to 100 iterations, but most of the time the ICM (Iterated Conditional Modes)
# loop will converge before reaching the 100th iteration.
# After setting the necessary parameters we can now call an instance of the
# class "TissueClassifierHMRF" and its method called "classify" and input the
# parameters defined above to perform the segmentation task.
#
# Now we plot the resulting segmentation.

t0 = time.time()

hmrf = TissueClassifierHMRF()
initial_segmentation, final_segmentation, PVE = hmrf.classify(t1, nclass, beta)

t1 = time.time()
total_time = t1 - t0
print(f"Total time: {total_time}")

fig = plt.figure()
a = fig.add_subplot(1, 2, 1)
img_ax = np.rot90(final_segmentation[..., 89])
imgplot = plt.imshow(img_ax)
a.axis("off")
a.set_title("Axial")
a = fig.add_subplot(1, 2, 2)
img_cor = np.rot90(final_segmentation[:, 128, :])
imgplot = plt.imshow(img_cor)
a.axis("off")
a.set_title("Coronal")
plt.savefig("final_seg.png", bbox_inches="tight", pad_inches=0)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Each tissue class is color coded separately, red for the WM, yellow for
# the GM and light blue for the CSF.
#
#
# And we will also have a look at the probability maps for each tissue class.

fig = plt.figure()
a = fig.add_subplot(1, 3, 1)
img_ax = np.rot90(PVE[..., 89, 0])
imgplot = plt.imshow(img_ax, cmap="gray")
a.axis("off")
a.set_title("CSF")
a = fig.add_subplot(1, 3, 2)
img_cor = np.rot90(PVE[:, :, 89, 1])
imgplot = plt.imshow(img_cor, cmap="gray")
a.axis("off")
a.set_title("Gray Matter")
a = fig.add_subplot(1, 3, 3)
img_cor = np.rot90(PVE[:, :, 89, 2])
imgplot = plt.imshow(img_cor, cmap="gray")
a.axis("off")
a.set_title("White Matter")
plt.savefig("probabilities.png", bbox_inches="tight", pad_inches=0)
plt.show()

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# These are the probability maps of each of the three tissue classes.
#
#
# References
# ----------
#
# .. footbibliography::
#
